#include "audioddc.h"

AudioDDC::AudioDDC()
{
    fft_ = NULL;
    local_oscillator_ = NULL;
    capture_data_ = NULL;
    captured_data_i_ = NULL;
    captured_data_q_ = NULL;
    captured_data_fft_dbm_ = NULL;

    lo_initialized_ = false;
    capture_initialized_ = false;
    waterfall_initialized_ = false;
    capture_max_size_ = 0;
}

AudioDDC::~AudioDDC()
{
    if (fft_) delete fft_;
    if (local_oscillator_) delete local_oscillator_;
    if (capture_data_) delete capture_data_;

    if (waterfall_initialized_) {
        delete[] captured_data_i_;
        delete[] captured_data_q_;
        delete[] captured_data_fft_dbm_;
    }
}

void AudioDDC::init_lo(int address, int memsize, double master_frequency, int phase_width, double init_frequency) {
    if (local_oscillator_) delete local_oscillator_;
    local_oscillator_ = new DDS(address, memsize, master_frequency, phase_width);
    local_oscillator_ -> set_frequency(init_frequency);
    lo_initialized_ = true;
}

void AudioDDC::init_capture(int address, int memsize, int capture_size) {
    if (capture_data_) delete capture_data_;
    capture_data_ = new Data_Capture_Dual(address, memsize, capture_size);
    capture_max_size_ = capture_size;
    capture_initialized_ = true;
}

void AudioDDC::init_waterfall(WaterfallPlot *waterfall_audio, int max_frequency, double data_range_min, double data_range_max, double span_init, int data_size, int waterfall_depth) {

    if ((lo_initialized_) && (capture_initialized_) && (capture_max_size_ >= data_size)) {
        data_size_ = data_size;

        if (fft_) delete fft_;
        fft_ = new fft_basic(data_size_, fftw_base::hann);

        waterfall_audio_ = waterfall_audio;
        waterfall_audio_ -> init(data_size_, waterfall_depth, max_frequency, data_range_min, data_range_max, 0, span_init, local_oscillator_->get_frequency());
        waterfall_audio_ -> set_min_zoom(0);
        waterfall_audio_ -> set_max_zoom(0);
        waterfall_audio_ -> set_wheel_action(1);

        QwtLinearColorMap waterfall_colors(Qt::black, Qt::magenta);
        waterfall_colors.addColorStop(0.2, Qt::blue);
        waterfall_colors.addColorStop(0.3, Qt::green);
        waterfall_colors.addColorStop(0.4, Qt::yellow);
        waterfall_colors.addColorStop(0.5, Qt::red);

        waterfall_audio_->set_color_map(waterfall_colors);

        captured_data_i_ = new double[data_size_];
        captured_data_q_ = new double[data_size_];
        captured_data_fft_dbm_ = new double[data_size_];

        waterfall_initialized_ = true;
    }
    else {
        cerr << "ERROR audioddc.cpp: run init_lo and init_capture before init_waterfall or check fft/waterfall data size with max data size of capture IP core " << endl;
        exit(1);
    }
}



double AudioDDC::set_rx_frequency(double rx_frequency) {
    if ((lo_initialized_) && (waterfall_initialized_)) {
        waterfall_audio_ -> set_center_frequency(rx_frequency);
        return local_oscillator_ -> set_frequency(rx_frequency);
    }
    else {
        cerr << "ERROR: audio_ddc.cpp: LO not initialized, call init_lo before setting the RX frequency" << endl;
        exit(1);
    }
}


void AudioDDC::capture_and_show(double gain_correction) {
    if (waterfall_initialized_) {

        QTime zeitmessung;
        zeitmessung.start();
        int capture_fft_time = 0;
        int display_spectrum_time = 0;

        // read data and do FFT
        int audio_cap_err = capture_data_ -> capture_and_read(captured_data_i_, captured_data_q_, data_size_);
        if (audio_cap_err != 0) cerr << "AUDIO CAPTURE ERROR CODE " << audio_cap_err << endl;

        fft_->complex_fft_dbm(captured_data_i_, captured_data_q_, captured_data_fft_dbm_);

        for (int i=0; i<data_size_; i++)
            captured_data_fft_dbm_[i] = captured_data_fft_dbm_[i] - gain_correction;

        capture_fft_time = zeitmessung.elapsed();

        // display FFT data
        if (waterfall_audio_->receiver_update_required() == false)
            waterfall_audio_->set_next_data_line(captured_data_fft_dbm_);

        display_spectrum_time = zeitmessung.elapsed();

        // do a frequency update, if necessary
        if (waterfall_audio_->receiver_update_required()) {
            local_oscillator_ -> set_frequency(waterfall_audio_->get_center_frequency());
            emit this->send_new_lo_frequency(waterfall_audio_->get_center_frequency());

            waterfall_audio_ ->reset_receiver_update();
            cerr << "audio receiver update, new frequency: " <<waterfall_audio_->get_center_frequency() << endl;
        }

        cerr << "Audio times: " << \
                "cap&fft: " << capture_fft_time << ", " << \
                "spectrum: " << display_spectrum_time - capture_fft_time << endl;


    }
    else {
        cerr << "ERROR: audio_ddc.cpp: run init_waterfall before calling capture_and_show" << endl;
        exit(1);
    }

}
