#ifndef AUDIODDC_H
#define AUDIODDC_H

#include "../waterfall/waterfallplot.h"
#include "fft/fft_basic.h"
#include "stipcores.h"

/** @brief Class for controlling a single Audio DDC
 *  - first call init functions init_lo, init_capture and init_waterfall (in this order) to set up the class before use
 *  - use other functions to set the RX frequency or display a new line in the waterfall
 */

class AudioDDC: public QObject
{
    Q_OBJECT
public:
    AudioDDC();
    ~AudioDDC();


    /** @brief initializes LO IP core, required to run before using the DDC!
     *  @param address IP core address (from Vivado)
     *  @param memsize IP core memory size in k  (from Vivado)
     *  @param master_frequency DDS master frequency in kHz (from Vivado)
     *  @param phase_width DDS phase width (from Vivado)
     *  @param init_frequency inital LO frequency in kHz */
    void init_lo(int address, int memsize, double master_frequency, int phase_width, double init_frequency);


    /** @brief initializes Dual Capture IP core for IQ data, required to run before using the DDC!
     *  @param address IP core address (from Vivado)
     *  @param memsize IP core memory size in k  (from Vivado)
     *  @param capture_size depth of capture memory in i/q sample pairs */
    void init_capture(int address, int memsize, int capture_size);


    /** @brief initializes the audio DDC, required to run after use!
      * @param *waterfall_audio pointer to waterfall display
      * @param max_frequency maximum frequency the waterfall can be scrolled to in kHz (e.g. 125000 for 125 MHz)
      * @param data_range_min minimum data value in waterfall
      * @param data_range_max maximum data value in waterfall
      * @param span_init initial frequency span of waterfall in kHz (= audio IQ signal's sample rate)
      * @param data_size size of data block to be captured and displayed
      * @param waterfall_depth depth of waterfall display  */
    void init_waterfall(WaterfallPlot *waterfall_audio, int max_frequency, double data_range_min, double data_range_max, double span_init, int data_size, int waterfall_depth);


    /** @brief sets LO frequency of DDC
      * @param rx_frequency in kHz
      * @return actually set frequency in kHz */
    double set_rx_frequency(double rx_frequency);

    /** @brief captures IQ data (after DDC) and shows them as a new line in the waterfall plot
        @param gain_correction correction factor in dB (will be subtracted from measured results in dbm) for the gain, that includes e.g. RF scaling gain, filter gain etc. */
    void capture_and_show(double gain_correction);

	/** @brief return the I&Q data for the last data capture (for debugging) 
		@param in[] array for writing the I data, array length = "data_size" (=size of data block to be captured and displayed)
		@param qu[] array for writing the Q data, array length = "data_size" (=size of data block to be captured and displayed)
	*/
    void get_captured_data(double in[], double qu[]) {
        if (waterfall_initialized_) {
            for (int i=0; i<data_size_; i++) {
                in[i] = captured_data_i_[i];
                qu[i] = captured_data_q_[i];
            }
        }
        else {
            cerr << "ERROR: audio_ddc.cpp: run init_waterfall before calling capture_and_show" << endl;
            exit(1);
        }
    }

	/** @brief selects the correct size of the audio receiver's waterfall */
    void correct_waterfall_size() {
        if (waterfall_initialized_)  waterfall_audio_->correct_size();
    }

signals:
	/** @brief send the current LO frequency (after it has been newly set, e.g. by a waterfall shift) 
	  * @param frequency LO frequency in kHz 	*/
    void send_new_lo_frequency(double frequency);
	
private:
    fft_basic           *fft_;
    DDS                 *local_oscillator_;		// Driver for LO
    Data_Capture_Dual   *capture_data_;
    WaterfallPlot       *waterfall_audio_;		// Waterfall Display Class

    double              *captured_data_i_;
    double              *captured_data_q_;
    double              *captured_data_fft_dbm_;

    int                 capture_max_size_;      // depth of capture ip core memory
    int                 data_size_;             // data block length for fft and waterfall

    bool                lo_initialized_;		// signal, which init functions have been called
    bool                capture_initialized_;
    bool                waterfall_initialized_;

};

#endif // AUDIODDC_H
