#include "audiorx.h"

AudioRX::AudioRX() {
    num_audio_ddcs_ = 2;

    ddc_selector_   = NULL;
    reset_module_ = NULL;

    capture_spectrum_out_ = NULL;
    spectrum_out_curve_ = NULL;
    spectrum_out_grid_ = NULL;
    fft_ = NULL;
    captured_data_i_ = NULL;
    captured_data_q_ = NULL;
    captured_data_fft_db_ = NULL;
    captured_data_fft_db_real_ = NULL;
    audio_scaler_i_ = NULL;
    audio_scaler_q_ = NULL;
    mux_controller_ = NULL;
    weaver_lo_      = NULL;

    max_scaling_db_ = 0;
    binaural_out_ = 0;

    audio_ddc_ = new AudioDDC*[num_audio_ddcs_];
    for (int i = 0; i < num_audio_ddcs_; i++)
        audio_ddc_[i] = new AudioDDC();

    connect(audio_ddc_[0],SIGNAL(send_new_lo_frequency(double)), this, SIGNAL(send_new_frequency_ddc0(double)));
    connect(audio_ddc_[1],SIGNAL(send_new_lo_frequency(double)), this, SIGNAL(send_new_frequency_ddc1(double)));

    reset_initialized_ = false;
    ddc_selector_initialized_ = false;
    spectrum_out_initialized_ = false;
    scaler_initalized_ = false;
    mux_controller_initialized_ = false;
    weaver_lo_initialized_ = false;

    ddc_out_active_ = new bool[num_audio_ddcs_];
    for (int i=0; i<num_audio_ddcs_; i++) ddc_out_active_[i] = false;

    heaphone_out_active_ = false;

}

AudioRX::~AudioRX() {
    if (ddc_selector_) delete ddc_selector_;
    if (reset_module_) delete reset_module_;

    if (capture_spectrum_out_) delete capture_spectrum_out_;
    if (spectrum_out_curve_) delete spectrum_out_curve_;
    if (spectrum_out_grid_) delete spectrum_out_grid_;
    if (fft_) delete fft_;

    if (spectrum_out_initialized_) {
        delete[] captured_data_i_;
        delete[] captured_data_q_;
        delete[] captured_data_fft_db_;
        delete[] captured_data_fft_db_real_;
    }

    if (audio_scaler_i_) delete audio_scaler_i_;
    if (audio_scaler_q_) delete audio_scaler_q_;
    if (mux_controller_) delete mux_controller_;
    if (weaver_lo_) delete weaver_lo_;

    for (int i = 0; i < num_audio_ddcs_; i++)
        delete audio_ddc_[i];

    delete[] audio_ddc_;
    delete[] ddc_out_active_;

}

void AudioRX::init_reset(int address, int memsize) {
    if (reset_module_) delete reset_module_;
    reset_module_ = new IP_Driver(address, memsize);
    reset_module_ -> write(0,0);
    reset_module_ -> write(0,0xFFFFFFFF);
    reset_initialized_ = true;
}


void AudioRX::init_ddc_selector(int address, int memsize) {
    if (ddc_selector_) delete ddc_selector_;
    ddc_selector_ = new IP_Driver(address, memsize);
    ddc_selector_initialized_ = true;
}


void AudioRX::init_lo(int address, int memsize, double master_frequency, int phase_width, double init_frequency, int ddc_number) {
    if ((ddc_number >= num_audio_ddcs_) || (ddc_number < 0)) {
        cerr << "ERROR audiorx.cpp: init_lo: invalid ddc number " << ddc_number << " specified" << endl;
        exit(1);
    }

    audio_ddc_[ddc_number] -> init_lo(address, memsize, master_frequency, phase_width, init_frequency);

}

void AudioRX::init_capture(int address, int memsize, int capture_size, int ddc_number) {
    if ((ddc_number >= num_audio_ddcs_) || (ddc_number < 0)) {
        cerr << "ERROR audiorx.cpp: init_capture: invalid ddc number " << ddc_number << " specified" << endl;
        exit(1);
    }

    audio_ddc_[ddc_number] -> init_capture(address, memsize, capture_size);

}

void AudioRX::init_waterfall(WaterfallPlot *waterfall_audio, int max_frequency, double data_range_min, double data_range_max, double span_init, int data_size, int waterfall_depth, int ddc_number) {
    if ((ddc_number >= num_audio_ddcs_) || (ddc_number < 0)) {
        cerr << "ERROR audiorx.cpp: init_waterfall: invalid ddc number " << ddc_number << " specified" << endl;
        exit(1);
    }

    audio_ddc_[ddc_number] -> init_waterfall(waterfall_audio, max_frequency, data_range_min, data_range_max, span_init, data_size, waterfall_depth);
}

void AudioRX::init_scaling(int i_address, int i_memsize, int q_address, int q_memsize, int max_scaling_db) {
    if (audio_scaler_i_) delete audio_scaler_i_;
    audio_scaler_i_ = new Signal_Scaler(i_address, i_memsize, max_scaling_db);

    if (audio_scaler_q_) delete audio_scaler_q_;
    audio_scaler_q_ = new Signal_Scaler(q_address, q_memsize, max_scaling_db);

    max_scaling_db_ = max_scaling_db;

    scaler_initalized_ = true;

    set_audio_scaling(0); // audio scaling default 0 dB

    // Timer for RF Scaler Update
    QTimer *timer_audio_scaler = new QTimer(this);
    connect(timer_audio_scaler, SIGNAL(timeout()), this, SLOT(scaler_timer_slot()));
    timer_audio_scaler->start(500);
}

void AudioRX::init_mux_controller(int address, int memsize) {
    if (mux_controller_) delete mux_controller_;
    mux_controller_ = new IP_Driver(address, memsize);
    mux_controller_initialized_ = true;
}

void AudioRX::init_weaver_lo(int address, int memsize, double master_frequency, int phase_width, double init_frequency) {
    if (weaver_lo_) delete weaver_lo_;
    weaver_lo_ = new DDS(address, memsize, master_frequency, phase_width);
    weaver_lo_ -> set_frequency(init_frequency);
    weaver_lo_initialized_ = true;
}


void AudioRX::get_ddc_frequency(double frequency, int ddc_number) {
    if ((ddc_number >= num_audio_ddcs_) || (ddc_number < 0)) {
        cerr << "ERROR audiorx.cpp: get_ddc_frequency: invalid ddc number " << ddc_number << " specified" << endl;
        exit(1);
    }

    audio_ddc_[ddc_number] -> set_rx_frequency(frequency);
    emit this -> send_true_frequency(frequency, ddc_number);
}

void AudioRX::capture_and_show(double gain_correction, int ddc_number) {
    if ((ddc_number >= num_audio_ddcs_) || (ddc_number < 0)) {
        cerr << "ERROR audiorx.cpp: capture_and_show: invalid ddc number " << ddc_number << " specified" << endl;
        exit(1);
    }

    audio_ddc_[ddc_number] -> capture_and_show(gain_correction);
    return;
}


void AudioRX::init_output_spectrum_display(int capture_address, int capture_memsize, int capture_size, QwtPlot *spectrum_out_plot) {
    capture_size_ = capture_size;

    // FFT variables
    if (fft_) delete fft_;
    fft_ = new fft_basic(capture_size_, fftw_base::hann);

    captured_data_i_ = new double[capture_size_];
    captured_data_q_ = new double[capture_size_];
    captured_data_fft_db_ = new double[capture_size_];
    captured_data_fft_db_real_ = new double[capture_size_/2];

    // define Qwt plot
    spectrum_out_plot_ = spectrum_out_plot;
    spectrum_out_plot_ -> setAxisScale(0,-100, +30, 20);

    // set Fonts for X11
    QFont axisfont = spectrum_out_plot_->axisFont(0); // read default font
    axisfont.setPointSize(8);
    spectrum_out_plot_ -> setAxisFont(0,axisfont);
    spectrum_out_plot_ -> setAxisFont(2,axisfont);
    spectrum_out_plot_ -> replot();

    // define a plot curve and attach it to the QwtPlot
    spectrum_out_curve_ = new QwtPlotCurve();
    spectrum_out_curve_ -> setPen(QPen(Qt::blue, 2));
    spectrum_out_curve_ -> attach(spectrum_out_plot_);

    // attach a grid to the QwtPlot
    spectrum_out_grid_ = new QwtPlotGrid;
    spectrum_out_grid_ -> attach(spectrum_out_plot_);

    // init capture module
    if (capture_spectrum_out_) delete capture_spectrum_out_;
    capture_spectrum_out_ = new Data_Capture_Dual(capture_address, capture_memsize, capture_size_);

    spectrum_out_initialized_ = true;
}


void AudioRX::capture_and_show_spectrum_out(){
    if (!spectrum_out_initialized_) {
        cerr << "ERROR audiorx.cpp, capture_and_show_spectrum_out(): initialization not done, run init_output_sepctrum( ) first" << endl;
        exit(1);
    }

    // read data and do FFT
    int cap_err = capture_spectrum_out_ -> capture_and_read(captured_data_i_, captured_data_q_, capture_size_);
    if (cap_err != 0) cerr << "CAPTURE ERROR CODE " << cap_err << endl;

    if (binaural_out_) {
        fft_ -> complex_fft_dbm(captured_data_i_, captured_data_q_, captured_data_fft_db_);

        // for (int i = 0; i < capture_size_; i++)
        //     captured_data_fft_db_[i] = captured_data_fft_db_[i] - gain_correction;	// use if correct level of output spectrum is desired

        double frequency_scale[capture_size_];    // generate frequency scale
        for (int i=0; i<capture_size_; i++)
            frequency_scale[i] = -24 + i*(48 / (double)(capture_size_));

        spectrum_out_curve_ -> setRawData(frequency_scale, captured_data_fft_db_, capture_size_); // display FFT data
        spectrum_out_plot_ -> replot();
    }
    else {
        fft_ -> real_fft_dbm(captured_data_i_, captured_data_fft_db_real_);
        double frequency_scale[capture_size_/2];    // generate frequency scale
        for (int i=0; i<capture_size_/2; i++)
            frequency_scale[i] = 0+ i*(24 / (double)(capture_size_/2));

        spectrum_out_curve_ -> setRawData(frequency_scale, captured_data_fft_db_real_, capture_size_/2); // display FFT data
        spectrum_out_plot_ -> replot();
    }

}



void AudioRX::reset() {
    if (!reset_initialized_) {
        cerr << "ERROR audiorx.cpp, reset(): tried to apply reset, but IP core was not initalized, call init_reset first" << endl;
        exit(1);
    }

    // apply reset
    reset_module_ -> write(0,2);
    reset_module_ -> write(0,1);
    cerr << "audio IQ reset executed" << endl;
}

void AudioRX::correct_size() {
    for (int i = 0; i < num_audio_ddcs_; i++)
        audio_ddc_[i] -> correct_waterfall_size();
}



void AudioRX::get_output_ddc_selection(int ddc_number, bool active) {
    ddc_out_active_[ddc_number] = active;

    int mux_config = 0;
    if (!ddc_out_active_[0] && !ddc_out_active_[1]) mux_config = 0;  // select  DDC 0
    if (ddc_out_active_[0] && !ddc_out_active_[1]) mux_config = 0;  // select  DDC 0
    if (!ddc_out_active_[0] && ddc_out_active_[1]) mux_config = 1;  // select  DDC 1
    if (ddc_out_active_[0] && ddc_out_active_[1]) mux_config = 2;  // select  DDC 0+1

    ddc_selector_ -> write(0x0, mux_config);
}

void AudioRX::get_output_properties(int bandwidth, bool inverted, bool binaural) {
    if (!mux_controller_initialized_ || !weaver_lo_initialized_) {
        cerr << "ERROR audiorx.cpp, set_output_properties: mux controller or weaver lo not initialized" << endl;
        exit(1);
    }

    int bandwidth_config = 0;

    switch (bandwidth) {
        case 0: bandwidth_config = 0;
                weaver_lo_ -> set_frequency(12);  // 12 kHz shift in weaver demodulator, maybe change to 11.5 if overlap is allowed (for better audio listening)
                break;

        case 1: bandwidth_config = 2;
                weaver_lo_ -> set_frequency(3.3);
                break;

        case 2: bandwidth_config = 3;
                weaver_lo_ -> set_frequency(1.5);
                break;
    }

    int config_value = bandwidth_config + (!inverted) * 4 + (!binaural) * 8;
    mux_controller_->write(0, config_value);

    binaural_out_ = binaural;
}


void AudioRX::get_output_scaler_setting(int scaling_value_db) {
    manual_scaler_setting_db_ = scaling_value_db;
    if (!auto_scaling_) set_audio_scaling(manual_scaler_setting_db_);
}

void AudioRX::get_output_scaler_auto_setting(bool auto_setting) {
    auto_scaling_ = auto_setting;
    if (!auto_scaling_) set_audio_scaling(manual_scaler_setting_db_);
}

void AudioRX::set_audio_scaling(int scaling_factor_db) {
    if (!scaler_initalized_) {
        cerr << "ERROR audiorx.cpp: tried to set scaling, but IP core was not initalized, call init_scaling first" << endl;
        exit(1);
    }

    // set scaling value
    audio_scaler_i_ -> set_scaling(scaling_factor_db);
    audio_scaler_q_ -> set_scaling(scaling_factor_db);

    // send scaling value to GUI
    emit this->send_scaling_value(scaling_factor_db);
}


void AudioRX::scaler_timer_slot() {
    // get max audio value since last timer step
    int audio_max_value = audio_scaler_i_ -> get_max_value();
    int audio_max_value_dbfs = (int) (20*log10((double)audio_max_value / 8388608));

    emit this -> send_audio_out_level(audio_max_value, audio_max_value_dbfs);
    audio_scaler_i_ -> reset_max_value();

    // Scale the signal automatically
    if (auto_scaling_) {

        if (audio_max_value_dbfs > -40) {
            // set to no amplification
            set_audio_scaling(0);
        }

        if (audio_max_value_dbfs < -60) {
            // set amplification
            int requested_scaling_db = (-audio_max_value_dbfs-10) / 6;
            if (requested_scaling_db <= max_scaling_db_)
                set_audio_scaling(requested_scaling_db );
            else
                set_audio_scaling(max_scaling_db_);
        }
    }

}


void AudioRX::get_audio_headphone_out(bool checkbox_status) {
    // When activating the audio output for headphone out, the sound card needs to be activated, e.g. by a dummy loopback command
    // either here or with external scripts 
    
    if (!heaphone_out_active_ && checkbox_status) {
        //activate sound card
        heaphone_out_active_ = true;
    }

    if (heaphone_out_active_ && !checkbox_status) {
        // deactivate sound card
        heaphone_out_active_ = false;
    }
    
}
