#ifndef AUDIORX_H
#define AUDIORX_H

#include "../waterfall/waterfallplot.h"
#include "stipcores.h"
#include "fft/fft_basic.h"
#include "fft/fft_avg_complex.h"
#include "audioddc.h"

#include <qwt_plot_curve.h>
#include <qwt_plot_grid.h>
#include <qwt_plot.h>

#include <QTimer>

/** @brief Toplevel class for an audio receiver with two audio DDCs with IQ Reset (DDC)
  * - call init functions before use: init_reset, init_ddc_selector and then init_lo, init_capture, init_waterfall (last 3 in this order)
  * - controls the functions of the two audio DDCs
  * - controls the audio reset, audio scaler, bandwidth and SSB/Weaver selection
  */
  
class AudioRX : public QObject
{
    Q_OBJECT

public:
    /** @brief constructor */
    AudioRX();

    /** @brief destructor */
    ~AudioRX();

    /** @brief initializes the reset IP core
     * @param address IP core address from Vivado
     * @param memsize memory space size in k from Vivado */
    void init_reset(int address, int memsize);


    /** @brief initializes the MUX controller for selecting the audio DDCs
      * @param address IP core adress from Vivado
      * @param memsize memory space size in k from Vivado */
    void init_ddc_selector(int address, int memsize);

    /** @brief initializes LO IP core of one DDC
     *  @param address IP core address (from Vivado)
     *  @param memsize IP core memory size in k  (from Vivado)
     *  @param master_frequency DDS master frequency in kHz (from Vivado)
     *  @param phase_width DDS phase width (from Vivado)
     *  @param init_frequency inital LO frequency in kHz
     *  @param ddc_number number of DDC (0 or 1)*/
    void init_lo(int address, int memsize, double master_frequency, int phase_width, double init_frequency, int ddc_number);


    /** @brief initializes Dual Capture IP core for IQ data of one DDC
     *  @param address IP core address (from Vivado)
     *  @param memsize IP core memory size in k  (from Vivado)
     *  @param capture_size depth of capture memory in i/q sample pairs
	 *  @param ddc_number number of DDC (0 or 1) */
    void init_capture(int address, int memsize, int capture_size, int ddc_number);


    /** @brief initializes the waterfall for one DDC
      * @param *waterfall_audio pointer to waterfall display
      * @param max_frequency maximum frequency the waterfall can be scrolled to in kHz (e.g. 125000 for 125 MHz)
      * @param data_range_min minimum data value in waterfall
      * @param data_range_max maximum data value in waterfall
      * @param span_init initial frequency span of waterfall in kHz (= audio IQ signal's sample rate)
      * @param data_size size of data block to be captured and displayed
      * @param waterfall_depth depth of waterfall display  
	  * @param ddc_number number of DDC (0 or 1) */
    void init_waterfall(WaterfallPlot *waterfall_audio, int max_frequency, double data_range_min, double data_range_max, double span_init, int data_size, int waterfall_depth, int ddc_number);

    /** @brief initializes the scaling IP cores for I and Q components
     * @param i_address I scaler IP core address from Vivado
     * @param i_memsize I scaler memory space size in k from Vivado
     * @param q_address Q scaler IP core address from Vivado
     * @param q_memsize Q scaler memory space size in k from Vivado
     * @param max_scaling_db maximum scaling value in dB (e.g. 48 means -48dB  to +48dB) */
	void init_scaling(int i_address, int i_memsize, int q_address, int q_memsize, int max_scaling_db);

    /** @brief initializes the MUX controller for bandwidth, spectrum inversion and binaural mode selection
      * @param address IP core adress from Vivado
      * @param memsize memory space size in k from Vivado */
    void init_mux_controller(int address, int memsize);

    /** @brief initializes Weaver LO
     *  @param address IP core address (from Vivado)
     *  @param memsize IP core memory size in k  (from Vivado)
     *  @param master_frequency DDS master frequency in kHz (from Vivado)
     *  @param phase_width DDS phase width (from Vivado) 
	 *  @param init_frequency inital weaver frequency in kHz */
    void init_weaver_lo(int address, int memsize, double master_frequency, int phase_width, double init_frequency);


    /** @brief updates waterfall: captures data, does FFT and draws one line for the audio DDC
      * @param gain_correction correction of gain for FFT power spectrum, apart from CPG (e.g. scaling/filter/mixer gain)
      * @param ddc_number number of audio ddc (0 or 1) */
    void capture_and_show(double gain_correction, int ddc_number);


    /** @brief shows the audio output spectrum before the audio IF to the PS or ADAU
      * @param capture_address address of dual capture IP core
      * @param capture_memsize
      * @param capture_size capture size of dual capture IP core
      * @param spectrum_out_plot pointer to QwtPlot, which will display the spectrum */
    void init_output_spectrum_display(int capture_address, int capture_memsize, int capture_size, QwtPlot *spectrum_out_plot);

    /** @brief shows the audio output spectrum before the audio IF to the PS or ADAU*/
    void capture_and_show_spectrum_out();


    /** @brief resets all audio DDCs and audio out interface, i.e. IQ paths (and sample rate conversion and audio interface) */
    void reset();

    /** @brief corrects size of the waterfall plots of all audio DDCs */
    void correct_size();

    /** @brief returns if DDC with specified number (0 to num_audio_ddcs_-1) currently outputs data  */
    bool get_ddc_output_active(int ddc_number) {return ddc_out_active_[ddc_number];}


signals:
	/** @brief sends the current gain value (in dB) of the audio scaler */
    void send_scaling_value(int true_scaler_value_db);
	
	/** @brief sends the current audio level (after the audio DDC) */
    void send_audio_out_level(int audio_level, int audio_level_dbfs);
	
	/** @brief sends the actually set LO frequency after a frequency update has been recevied by slot get_ddc_frequency */
    void send_true_frequency(double frequency, int ddc);
	
	/** @brief sends the current LO frequency for DDC 0 (when it has been newly set in the DDC, e.g. originating in a waterfall shift)*/
    void send_new_frequency_ddc0(double frequency);
	
	/** @brief sends the current LO frequency for DDC 1 (when it has been newly set in the DDC, e.g. originating in a waterfall shift)*/
    void send_new_frequency_ddc1(double frequency);

	
public slots:
	/** @brief receives request for changing the audio processing properties
	  * @param bandwidth 0: wideband, 1: mediumband, 2: narrowband
	  * @param inverted spectrum inversion (other sideband)
	  * @param binaural	binaural mode, i.e. stereo I/Q  */
    void get_output_properties(int bandwidth, bool inverted, bool binaural);
	
	/** @brief receives request for changing the audio DDC's LO freqency
	  * @param frequency LO frequency in kHz
	  * @param ddc_number  */
    void get_ddc_frequency(double frequency, int ddc_number);
	
	/** @brief receives request for selection, which DDCs are active for outputting audio data */
    void get_output_ddc_selection(int ddc_number, bool active);
	
	/** @brief receives request to change the audio scaler gain setting*/
    void get_output_scaler_setting(int scaling_value_db);
	
	/** @brief receives request to change the automatic mode of the audio scaler*/
    void get_output_scaler_auto_setting(bool auto_setting);
	
	/** @brief receives timer signal (every 500ms) to automatically update the scaler setting and sending the current audio level (by calling signal "send_audio_out_level") */
    void scaler_timer_slot();

    /** @brief receive audio out chack box status */
    void get_audio_headphone_out(bool checkbox_status);

private:

    void set_audio_scaling(int scaling_factor_db);  	// low level function, that sets gain values for the scaler

    IP_Driver           *ddc_selector_;
    IP_Driver           *reset_module_;
    AudioDDC            **audio_ddc_;

    Data_Capture_Dual   *capture_spectrum_out_;
    QwtPlot             *spectrum_out_plot_;
    QwtPlotCurve        *spectrum_out_curve_;
    QwtPlotGrid         *spectrum_out_grid_;
    fft_basic           *fft_;
    double              *captured_data_i_;
    double              *captured_data_q_;
    double              *captured_data_fft_db_;
    double              *captured_data_fft_db_real_;
    int                 capture_size_;

    int                 num_audio_ddcs_;				// current implementation supports only 2 DDCs, i.e. num_audio_ddcs_ = 2

    bool                reset_initialized_;				// signal, which init functions have been called
    bool                ddc_selector_initialized_;
    bool                spectrum_out_initialized_;
    bool                scaler_initalized_;		
    bool                mux_controller_initialized_;
    bool                weaver_lo_initialized_;
	
    bool                *ddc_out_active_;

    Signal_Scaler       *audio_scaler_i_;
    Signal_Scaler       *audio_scaler_q_;
    IP_Driver           *mux_controller_;
    DDS                 *weaver_lo_;

    bool                auto_scaling_;
    int                 manual_scaler_setting_db_;
    int                 max_scaling_db_;

    bool                binaural_out_;
    bool                heaphone_out_active_;

};

#endif // AUDIORX_H
