#include "fftw_base.h"
#include <math.h>

using namespace std;

fftw_base::fftw_base(int n, t_fft_window fft_window) {

    n_ = n;
	resistor_ref_ = 50;
    adc_full_scale_voltage_ = 2;
    num_adc_bits_ = 16;

    in_ = (fftw_complex*) fftw_malloc(sizeof(fftw_complex)*n_);
    out_ = (fftw_complex*) fftw_malloc(sizeof(fftw_complex)*n_);

    my_fftw_plan_ = new fftw_plan();
    *my_fftw_plan_ = fftw_plan_dft_1d(n_, in_, out_, FFTW_FORWARD, FFTW_ESTIMATE);


	// window function

    fft_window_ = fft_window;

	window_weights_ = new double[n];

    if (fft_window_ == no) {
        for (int i = 0; i < n_; i++)
            window_weights_[i] = 1; // no window

        coherent_gain_ = 0;
        cerr << " construktor no window" << endl;
    }

    if (fft_window_ == hann) {
		for (int i = 0; i < n_; i++) 
			window_weights_[i] = 0.5*(1 - cos( (2*M_PI*i) /(n_-1)) ); // hann window (from wikipedia)

        coherent_gain_ = 6.02;
     }

    if (fft_window_ == flat) {
		double a0 = 0.21557895;
		double a1 = 0.41663158;
		double a2 = 0.277263158;
        double a3 = 0.083578947;
		double a4 = 0.006947368;

		for (int i = 0; i < n_; i++) 
			window_weights_[i] = a0 - a1 * cos((2*M_PI*i)/(n_-1)) + a2 * cos((4*M_PI*i)/(n_-1)) - a3 * cos((6*M_PI*i)/(n_-1)) + a4 * cos((8*M_PI*i)/(n_-1)); // flattop window (matlab)

        coherent_gain_ = 13.33;
	}

    data_out_mag_complex_temp_ = new double[n_];

}


fftw_base::~fftw_base() {
	delete[] window_weights_;
    delete[] data_out_mag_complex_temp_;

    fftw_destroy_plan(*my_fftw_plan_);
    fftw_free(in_);
    fftw_free(out_);
}


void fftw_base::real_fft_linear_power(double data_in[], double data_out_mag[])
{
	// copy data
	if (fft_window_ == no) {
		for (int i=0; i < n_; i++) {
            in_[i][0] = data_in[i];
            in_[i][1] = 0;
		}
	}
	else {
		for (int i=0; i < n_; i++) {
            in_[i][0] = window_weights_[i] * data_in[i];
            in_[i][1] = 0;
		}
	}

	// calculate fft
    fftw_execute(*my_fftw_plan_);

	// convert N point fft output to squared N/2 point (possible because of symmetry)
	// calculate voltage spectrum, linear scale
    data_out_mag[0] = (1.0/n_) * (adc_full_scale_voltage_ / pow(2,num_adc_bits_)) * sqrt(out_[0][0] * out_[0][0] + out_[0][1] * out_[0][1]);
    
	for (int i = 1; i < n_/2; i++)
        data_out_mag[i] = 2 * (1.0/n_) * (adc_full_scale_voltage_ / pow(2,num_adc_bits_)) * sqrt( out_[i][0]*out_[i][0] + out_[i][1]*out_[i][1] );

		
	// calculate power spectrum, linear scale
    for (int i = 0; i < n_/2; i++)
        data_out_mag[i] = data_out_mag[i] * data_out_mag[i];


}

void fftw_base::linear2dbm_real(double data_out_mag[]) {
	// calculate power spectrum magnitude in dbm, i.e. log scale
	for (int i = 0; i < n_/2; i++)
        data_out_mag[i] = 10* log10 ( data_out_mag[i] / (resistor_ref_*0.001) );
}

void fftw_base::correct_coherent_gain_real(double data_out_mag[]) {
	// correct coherent power gain
	for (int i = 0; i < n_/2; i++)
		data_out_mag[i] += coherent_gain_;
}


void fftw_base::complex_fft_linear_power(double data_in_real[], double data_in_imag[], double data_out_mag[])
{
	// copy data
	if (fft_window_ == no) {
		for (int i=0; i < n_; i++) {
            in_[i][0] = data_in_real[i];
            in_[i][1] = data_in_imag[i];
		}
	}
	else {
		for (int i=0; i < n_; i++) {
            in_[i][0] = window_weights_[i] * data_in_real[i];
            in_[i][1] = window_weights_[i] * data_in_imag[i];
		}
	}


	// calculate fft
    fftw_execute(*my_fftw_plan_);


	// calculate voltage spectrum, linear scale
    for (int i = 0; i < n_; i++)
        data_out_mag[i] = (1.0/n_) * (adc_full_scale_voltage_ / pow(2,num_adc_bits_)) * sqrt( out_[i][0]*out_[i][0] + out_[i][1]*out_[i][1] );


	// calculate power spectrum, linear scale
    for (int i = 0; i < n_; i++)
        data_out_mag[i] = data_out_mag[i] * data_out_mag[i];


    // reordering of the bins to reflect the correct (IQ) spectrum
    for (int i=0; i< n_; i++)
        data_out_mag_complex_temp_[i] = data_out_mag[i];

    for (int i = 0; i < n_/2; i++) {
        data_out_mag[i] = data_out_mag_complex_temp_[(n_/2)-1-i];
        data_out_mag[i+(n_/2)] = data_out_mag_complex_temp_[n_-1-i];
    }

}

void fftw_base::linear2dbm_complex(double data_out_mag[]) {
	for (int i = 0; i < n_; i++)
        data_out_mag[i] = 10* log10 ( data_out_mag[i] / (resistor_ref_*0.001) );
}

void fftw_base::correct_coherent_gain_complex(double data_out_mag[]) {
	// correct coherent power gain
	for (int i = 0; i < n_; i++)
		data_out_mag[i] += coherent_gain_;
}

