#ifndef FFTW_BASE_H
#define FFTW_BASE_H

#include <iostream>
#include<fftw3.h>


/** @brief Base class for FFT for time domain voltage input data from an ADC (wrapped FFTW) */
class fftw_base
{

public:
	/** @brief window types */
    enum t_fft_window {no, hann, flat};
	
    /** @brief constructor, allocates memory, sets parameter for magnitude calculation and windowing coefficients
      * @param n FFT size an length of input data
      * @param fft_window FFT window to use */
    fftw_base(int n, t_fft_window fft_window);

    /** @brief destructor */
    virtual ~fftw_base();


protected:

    /** @brief FFT size */
    int n_;

    /** @brief resistance in Ohms of a resistor over which the input voltage occurs */
    double resistor_ref_;

    /** @brief full scale voltage of ADC in Volts */
    double adc_full_scale_voltage_;

    /** @brief number of ADC bits */
    int num_adc_bits_;

    /** @brief coherent power gain of selected window */
    double coherent_gain_;

    /** @brief selected window */
    t_fft_window fft_window_;

    /** @brief pointer to array of window weight coefficients */
    double *window_weights_;

    /** @brief FFTW specific module for FFT calculation */
    fftw_plan *my_fftw_plan_;

    /** @brief FFTW specific buffer for FFT input data */
    fftw_complex *in_;

    /** @brief FFTW specific buffer for FFT output data */
    fftw_complex *out_;
	
    /** @brief temporary array required when reordering the bins of a complex FFT */
    double *data_out_mag_complex_temp_;



    /** @brief performs FFT on real input values, including windowing, no CPG correction
      * @param data_in[] array containing n real input time domain samples
      * @param data_out_mag[] array for the n/2 FFT bins of the linear scaled power spectrum */
    void real_fft_linear_power(double data_in[], double data_out_mag[]);

    /** @brief performs FFT on complex input values, including windowing, no CPG correction
      * @param data_in_real[] array containing n input time domain samples, real part
      * @param data_in_imag[] array containing n input time domain samples, imaginary part
      * @param data_out_mag[] array for the n FFT bins of the linear scaled power spectrum */
    void complex_fft_linear_power(double data_in_real[], double data_in_imag[], double data_out_mag[]);


    /** @brief converts a linear scaled power spectrum of real input values into a power spectrum in dBm (same as "linear2dbm_complex" but with n/2 values)
      * @param data_out_mag[] array containing n/2 bins of a linear power spectrum, that will be converted to dBm in situ  */
    void linear2dbm_real(double data_out_mag[]);

    /** @brief converts a linear scaled power spectrum of complex input values into a power spectrum in dBm (same as "linear2dbm_real" but with n values)
      * @param data_out_mag[] array containing n bins of a linear power spectrum, that will be converted to dBm in situ  */
    void linear2dbm_complex(double data_out_mag[]);


    /** @brief corrects the coherent power gain (caused by windowing) of a power spectrum in dBm of real input values (n/2 samples)
      * @param data_out_mag[] array containing n/2 bins of a power spectrum in dBm, that will be corrected in situ  */
    void correct_coherent_gain_real(double data_out_mag[]);

    /** @brief corrects the coherent power gain (caused by windowing) of a power spectrum in dBm of complex input values (n samples)
      * @param data_out_mag[] array containing n bins of a power spectrum in dBm, that will be corrected in situ  */
    void correct_coherent_gain_complex(double data_out_mag[]);
};

#endif // FFTW_BASE_H
