#ifndef FRONTEND_H
#define FRONTEND_H

#include <QObject>
#include <math.h>
#include "stipcores.h"
#include <QTimer>

/** @brief Class for controlling the SDR frontend, i.e. the input select relais (GPIO IP core) and the rf scaler */
class Frontend : public QObject
{
    Q_OBJECT
public:
	/** @brief constructor */
    explicit Frontend(QObject *parent = 0);
	
	/** @brief destructor */
    ~Frontend();

	/** @brief initialization function for the class
	  * @param signalscaler_adr IP core address for RF signal scaler (from Vivado)
	  * @param signalscaler_memsize IP core memory size in k  (from Vivado)
	  * @param signalscaler_max maximum scaling value of scaler IP core in dB (e.g. 48 means -48 dB to +48 dB) 
	  * @param relais_controller_adr IP core address for GPIO IP core for relais control (from Vivado)
	  * @param relais_controller_size IP core memory size in k  (from Vivado) */	  
    void init(int signalscaler_adr, int signalscaler_memsize, int signalscaler_max, int relais_controller_adr, int relais_controller_size);

signals:
	/** @brief sends the scaling value after the scaler has been set newly
	    @param true_scaler_value_db scaling value in dB*/
    void send_scaling_value(int true_scaler_value_db);
	
	/** @brief sends the current RF level
		@param rf_level as linear value
		@param rf_level_dbfs in dBFS */
    void send_rf_level(int rf_level, int rf_level_dbfs);

public slots:

	/** @brief receives request to set the scaler
	    @param scaler_value_db requested scaler setting in dB */
    void get_scaler_setting(int scaler_value_db);

	/** @brief receives request to change the scaler mode
	    @param auto_setting manually/automatically */
    void get_auto_setting(bool auto_setting);
	
	/** @brief receives request to change the relais setting
	    @param value 0: AAF input, 1: 70cm input, 2: direct */
    void get_relais_setting(int value);
	
	/** @brief timer slot, send the current RF level and (if activated) automatically sets the scaler */
    void timer_slot();

private:
	/** @brief sets scaler value of RF scaler
	  * @param scaling_value_db */	
    void set_rf_scaler(int scaling_value_db) ;

    Signal_Scaler   *scaler_rf_;				// scaler driver
    IP_Driver       *relais_controller_;		// relais driver (GPIO)

    int             manual_scaler_setting_db_;
    bool            auto_scaling_;

    int             max_scaling_db_;
};

#endif // FRONTEND_H
