#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <fstream>
#include <QMainWindow>
#include "stipcores.h"
#include <qwt_plot_curve.h>
#include <qwt_plot_grid.h>
#include <qwt_color_map.h>
#include "audio_rx/audiorx.h"
#include "waterfall_rx/waterfallddc.h"
#include "testgenerator.h"
#include "dialogdebugmonitor.h"
#include "frontend.h"

#include <fft/fft_avg_complex.h>
#include <fft/fft_avg_real.h>

#include <QMouseEvent>


namespace Ui {
class MainWindow;
}

/** @brief Main class for controlling the Panoradio SDR */
class MainWindow : public QMainWindow
{
    Q_OBJECT
    
public:
    explicit MainWindow(QWidget *parent = 0);
    ~MainWindow();

   
signals:
    // Frontend Signals
	/** @brief sends setting of the frontend scaler slider */
    void send_frontend_scaler_setting(int scaler_value_db);
	
	/** @brief sends selection status of automatic scaling mode  */
    void send_frontend_auto_setting(bool auto_setting);
		
    //void send_frontend_relais_setting(int value);

	
    // Debug Signals
	/** @brief sends new debug frequency to the testgenerator  
	  * @param frequency in kHz */
    void send_debug_single(double frequency);
	
	/** @brief activates frequency sweep and sends parameters to the testgenerator
	  * @param start_f start frequency in kHz 
	  * @param stop_f stop frequency in kHz 
	  * @param step_f step frequency in kHz  */
    void send_debug_sweep(double start_f, double stop_f, double step_f);
    //void send_debug_sweep_stopped(bool stopped);

	
    // Audio Signals
	/** @brief sends request for new audio frequency (in kHz) to audio RX  */
    void send_audio_rx_frequency(double frequency, int ddc_number);
	
	/** @brief sends request for activating or deactivating an audio DDC to audio RX  */
    void send_audio_ddc_selection(int ddc_number, bool active);
	
	/** @brief sends request for changing the audio processing properties to audio RX
	* @param bandwidth 0: wideband, 1: mediumband, 2: narrowband
	* @param inverted spectrum inversion (other sideband)
	* @param binaural	binaural mode, i.e. stereo I/Q  */
    void send_audio_output_properties(int bandwidth, bool inverted, bool binaural);
	
	/** @brief sends request to change the audio scaler gain setting (after slider has been moved) to audio RX */
    void send_audio_output_scaler_setting (int scaling_value_db);
	
	/** @brief sends request to change the automatic mode of the audio scaler to audio RX */
    void send_audio_output_scaler_auto_setting(bool autos_setting);


private slots:

	/** @brief initialize and start the radio  */
    void on_connect_pushButton_clicked();
	
	/** @brief initialize and start the radio  */
    void update_waterfall();

    
	// Frontend Slots
	/** @brief new RF level received -> display in thermo widget */
    void get_frontend_rf_value(int rf_level, int rf_level_dbfs);
    	
	/** @brief new scaling value set -> display in label and set gain_scaler_rf (for waterfalls) */
	void get_frontend_scaling_value(int true_scaler_value_db);
	
	/** @brief scaler for RF frontend moved -> send value in dB to frontend class */
    void on_rfscaling_horizontalSlider_sliderMoved(int position);
	
	/** @brief automatic check box clicked -> send value to frontend class and enable or disable slider	*/
    void on_checkBox_rf_scale_auto_clicked(bool checked);

	
	
    // Debug Slots
	/** @brief single frequency selected -> read frequency from lineEdit and send to Testgenerator */
    void debug_single_selected();
	
	/** @brief sweep frequency selected -> read sweep properties and send to Testgenerator */
    void debug_sweep_selected();
	
	/** @brief new frequency set -> show in GUI */
    void get_debug_true_frequency(double frequency);
	
	/** @brief stop radio and start debug mode */
    void on_debugmonitor_pushButton_clicked();
	
	/** @brief debug function for manually correting the waterfall plot sizes */
    void correct_plots();
	
	/** @brief debug function for manually applying a hardware reset of the waterfall DDCs*/
    void waterfall_reset();
	
	/** @brief debug function for manually applying a hardware reset of the audio DDCs */
    void audio_reset();

	

    //Audio Slots
	/** @brief new audio level received -> display in thermo widget */
    void get_audio_out_level(int audio_level, int audio_level_dbfs);
	
	/** @brief sends current audio output properties from GUI (bw, spectrum inv., binaural) to audio rx */
    void get_audio_properties();
	
	/** @brief activate/deactivate audio DDC 0 */
    void on_select_audio_ddc0_Button_clicked();
	
	/** @brief activate/deactivate audio DDC 1 */
    void on_select_audio_ddc1_Button_clicked();
	
	/** @brief frequency of audio DDC 0 in line edit changed -> send new freq to audio RX/DDC */
	void on_audio_ddc0_freq_lineEdit_textChanged();
	
	/** @brief frequency of audio DDC 1 in line edit changed -> send new freq to audio RX/DDC */
    void on_audio_ddc1_freq_lineEdit_textChanged();
	
	/** @brief right click in waterfall window -> sends the frequency for tuning the audio RX
      * @param frequency in kHz */
    void get_audio_frequency_clicked(double frequency);
	
	/** @brief frequency change by shift in the audio waterfall 0 -> update frequency label */
    void get_audio_ddc0_new_frequency(double frequency);
	
	/** @brief frequency change by shift in the audio waterfall 1 -> update frequency label */	
    void get_audio_ddc1_new_frequency(double frequency);
	
	
	/** @brief scaler moved -> send value in dB to audio scaler */
    void on_audioscaling_horizontalSlider_sliderMoved(int position);
	
	/** @brief automatic scaling check box clicked -> send value to audio rx and enable or disable slider */
    void on_checkBox_audio_scale_auto_clicked(bool checked);
	
	/** @brief new scaling value set -> display in label */
    void get_audio_scaling_value(int true_scaler_value_db);

	

	
    // Waterfall settings
	/** @brief update rate changed from GUI -> set timer for waterfall update accordingly */
    void on_update_rate_comboBox_currentIndexChanged(int index);
	
	/** @brief FFT averaging changed from GUI -> send number to waterfall DDCs */
    void on_avg_ffts_comboBox_currentIndexChanged(const QString &arg1);
	
	/** @brief color map for waterfall changed from GUI > calculate new QwtLinearColorMap and set them for the waterfalls */
    void update_waterfall_color_map();



    // low level debug stuff
	/** @brief starts/stops recording data to a file */
    void on_record_pushButton_clicked();


private:
    Ui::MainWindow *ui;
    DialogDebugMonitor *ui_debug_monitor;

	/** @brief update the display of the bandwidth bars in the audio waterfalls */
    void 			set_bandwidth_bars();

	AD9467          *adc;
	AudioRX         *audio_rx_;
    Frontend        *rx_frontend_;			// SDR Frontend Hardware Control	
    Testgenerator   *testgenerator_;		// SDR Hardware Testgen. Control
    IP_Driver       *iq_reset;				// reset module (hardware reset) for waterfall DDC

    WaterfallDDC    *waterfall_ddc_;		// Main Waterfall RX
    WaterfallDDC    *waterfall_1a_ddc_;		// Left Waterfall RX
    WaterfallDDC    *waterfall_1b_ddc_;		// Right Waterfall RX

	QTimer 			*timer_;					// timer for updating the waterfall

    double          power_level_min_;  		// lowest and highest power levels in waterfalls, set in constructor
    double          power_level_max_;
    double          gain_mixer;             // gain in db introduced by the mixer
    int             gain_scaler_rf;         // gain in db introduced by the rf scaler
    double          gain_cic;               // gain in db introduced by the CIC filter
    double          gain_fir;               // gain in db introduced by the FIR filter
    double          gain_fir_audio;         // gain in db introduced by the audio FIR filter

    bool 			debug_monitor_activated_;	// radio in debug mode, i.e. normal radio operation stopped

    fstream         record_file_;			// file recording variables
    bool            currently_recording_;	// in recording mode (writes FFT result of main waterfall in a file)
};

#endif // MAINWINDOW_H
