#ifndef TESTGENERATOR_H
#define TESTGENERATOR_H

#include <QObject>
#include "stipcores.h"
#include <QTimer>

/** @brief Class for controlling the RF test generator and selecting the input source (RF/ADC, test generator or step/impulse stimulus) */
class Testgenerator : public QObject
{
    Q_OBJECT
public:
	/** @brief constructor */
    explicit Testgenerator(QObject *parent = 0);
	
	/** @brief destructor */
    ~Testgenerator();

	/** @brief initialization function for the input mux GPIO
	* @param address IP core address for RF signal scaler (from Vivado)
	* @param memsize IP core memory size in k  (from Vivado)
	* @param init_value  */	
    void init_input_mux(int address, int memsize,int init_value);
	
    /** @brief initializes test generator IP core
     *  @param address IP core address (from Vivado)
     *  @param memsize IP core memory size in k  (from Vivado)
     *  @param master_frequency DDS master frequency in kHz (from Vivado)
     *  @param phase_width DDS phase width (from Vivado) */	
    void init_generator(int address, int memsize, double master_frequency, int phase_width);

signals:
	/** @brief send the current test frequency (after it has been newly set) 
	  * @param true_frequency test frequency in kHz 	*/
    void send_true_frequency(double true_frequency);

public slots:
	/** @brief request a change of the input source  
	  * @param source 0: ADC/RF (normal), 1: sine(RF) test generator, 2: step/impulse stimulus */
    void get_signal_source(int source);
	
	/** @brief request a change of the testgenerator to single frequency mode
	  * @param frequency in kHz for single freuqency mode */
    void get_single(double frequency);
	
	/** @brief request a change of the testgenerator to sweep mode
	  * @param start_f start frequency in kHz 
	  * @param stop_f stop frequency in kHz 
	  * @param step_f step frequency in kHz  */
    void get_sweep(double start_f, double stop_f, double step_f);
	
	/** @brief requests halt or continue for the sweep frequency */
    void get_sweep_stopped(bool stopped);

private slots:
	/** @brief increments the sweep frequency (timer slot)*/
    void sweep_timer_slot();

private:
    DDS         *test_generator_;			// RF sine generator for debugging
    double      test_sweep_start_freq_;
    double      test_sweep_stop_freq_;
    double      test_sweep_step_freq_;
    double      test_sweep_frequency_;
    bool        sweep_active_;
    bool        sweep_stopped_;

    IP_Driver   *input_mux_ctrl_;			// MUX controller (GPIO) for selecting the input signal
};

#endif // TESTGENERATOR_H
