#ifndef RASTERDATA_H
#define RASTERDATA_H

#include <iostream>
#include "fstream"

using namespace std;

static int pixel_cntr_x;
static int pixel_cntr_y;


/** @brief Implementation of the abstract QwtRasterData Class,
 *  - unfortunately QwtRasterData is not designed for plots, where each data is mapped to exactly one pixel
 *  - some dirty hacks were neccessary to implement the desired behaviour  */
class RasterData: public QwtRasterData
{
    double **data_array_;

    int data_array_xsize_;      // size of the data array to be displayed (fixed for one instance)
    int data_array_ysize_;

    double data_array_min_;     // minimum and maximum data value in the data array
    double data_array_max_;

    double xaxis_min_;          // upper and lower value for x and y axis (axis scale, not pixels)
    double yaxis_min_;
    double xaxis_max_;
    double yaxis_max_;

    double axis2index_x_;    	// number of points on axis for one data point(one array index)
    double axis2index_y_;

    double background_value_;   // value, that corresponds to the plot background and therefore its color


public:
    /** @brief constructor: only sets array size; background value, axis and data range are set to default values and need to be changed by their functions
     *  @param data_array_xsize width of data array
     *  @param data_array_ysize height of data array  */
    RasterData(int data_array_xsize, int data_array_ysize): QwtRasterData()
    {
        // set data range and plot axis
        set_data_range(0, 1);
        set_axis(0, 0, 10, 10);

        // allocate array, that contains data to be displayed
        data_array_xsize_ = data_array_xsize;   // size is fixed for one instance!
        data_array_ysize_ = data_array_ysize;

        background_value_ = -200; // set a default value for the background

        data_array_ = NULL;
    }


    ~RasterData() {        
    }


    virtual QwtRasterData *copy() const
    {
        RasterData *clone = new RasterData(data_array_xsize_, data_array_ysize_);
        clone -> set_data_range(data_array_min_, data_array_max_);
        clone -> set_axis(xaxis_min_, yaxis_min_, xaxis_max_, yaxis_max_);
        clone -> set_data(data_array_);
        clone -> set_background_value(background_value_);

        return clone;
    }

    /** @brief implemented method of abstract QwtRasterData, reads min and max data value
      * @returns QwtDoubleInterval min and max value as QwtDoubleInterval  */
    virtual QwtDoubleInterval range() const
    {
        return QwtDoubleInterval(data_array_min_, data_array_max_);
    }

    /** @brief implemented method of abstract QwtRasterData,
      * @returns value from 2D array to build the displayed picture from it    */
    virtual double value(double x, double y) const
    {
        int x_idx = (int) ( ((x - xaxis_min_) / axis2index_x_ ) + 0.5 );
        int y_idx = (int) ( (y - yaxis_min_)  / axis2index_y_ );

        double datum = 0;

        if (x_idx < data_array_xsize_)       // this "if" is required. beacuse the Spectrogram in Qwt has some bug: after changing the boundaries "x" goes beyond the boundingRect !?
            datum = data_array_[y_idx][x_idx];  // x and y exchanged to pivot the plot
        else
            datum = background_value_;

        if ((y - yaxis_min_) <= 1e-6)
            pixel_cntr_x++;

        if ((x - xaxis_min_) <= 1e-6)
            pixel_cntr_y++;

        return datum;
    }


        // ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        //  Additional methods
        // ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        /** @brief sets min and max occuring value in data array
          * @param data_array_min minimum data value in 2D array
          * @param data_array_max maximum data value in 2D array     */
        void set_data_range(double data_array_min, double data_array_max) {
            data_array_min_ = data_array_min;
            data_array_max_ = data_array_max;
        }


        /** @brief sets min and max axis values for the painting area (BoundingRect) in axis scale coordinates (not pixels!), defines the axis scales  */
        void set_axis(double xaxis_min, double yaxis_min, double xaxis_max, double yaxis_max) {
            if ((xaxis_min < 0) || (xaxis_max < 0)) cerr << "rasterdata.h: negative Plot Scale does not work! (xaxis_min=" <<xaxis_min << ",xaxis_max= " << xaxis_max << ")" << endl;

            xaxis_min_ = xaxis_min;
            yaxis_min_ = yaxis_min;
            xaxis_max_ = xaxis_max;
            yaxis_max_ = yaxis_max;

            setBoundingRect(QwtDoubleRect(xaxis_min_, yaxis_min_, xaxis_max_, yaxis_max_));

            axis2index_x_ = (xaxis_max_ - xaxis_min_) / (data_array_xsize_-1);
            axis2index_y_ = (yaxis_max_ - yaxis_min_) / (data_array_ysize_ -1);
        }


        /** @brief sets data array to be displayed
          * @**array pointer to 2d array, call by reference!   */
        void set_data(double **array)  {
            data_array_ = array;
        }

        /** @brief resets pixel counter, required to check for the plot size in pixels */
        void reset_pixel_counter() {
            pixel_cntr_x = 0;
            pixel_cntr_y = 0;
        }

        /** @brief reads pixel counter (x) , required to check for the plot size (x) in pixels */
        int get_pixels_x() {
            if (pixel_cntr_x == 0) cerr << "rasterdata.h: ATTENTION: x Pixel counter does not seem to work properly" << endl << endl;
            return pixel_cntr_x;
        }

        /** @brief reads pixel counter (y) , required to check for the plot size (y) in pixels */
        int get_pixels_y() {
            if (pixel_cntr_y == 0) cerr << "rasterdata.h: ATTENTION: y Pixel counter does not seem to work properly" << endl << endl;
            return pixel_cntr_y;
        }

        /** @brief sets value for background areas, required to fill rightmost area (bug workaround) */
        void set_background_value(double background_value) {background_value_ = background_value;}

		/** @brief function for debugging, reads class members  */
        void read_stuff() {
            cerr << " #################################" << endl << "rasterdata.h: auslesen_fuck(), Objektadresse: " << this << endl;
            cerr << "xaxis_max_ " << xaxis_max_ << ", Inhalt : "<< xaxis_max_ << endl;
            cerr << "xaxis_min_ " << xaxis_min_ <<", Inhalt : "<< xaxis_min_ <<  endl;
            cerr << "data_array_xsize_ " << data_array_xsize_ <<  endl;
            cerr << "axis2index_x_ " << axis2index_x_ <<", Inhalt : "<< axis2index_x_ <<  endl;
            cerr << "data_array_ "<< data_array_ << " &= "<<&data_array_<< endl;
            cerr << "Adresse der Klasse " << this << endl;
            cerr << "##################################" << endl << endl;
        }




};

#endif // RASTERDATA_H
