#include "waterfallplot.h"

WaterfallPlot::WaterfallPlot(QWidget *parent) :
    QwtPlot(parent)
{
    waterfall_data_ = NULL;
    spectrogram_plot_ = NULL;
    raster_data_ = NULL;
    receiver_update_required_ = false;
    wheel_action_ = 0;
}


WaterfallPlot::~WaterfallPlot() {
    if (spectrogram_plot_)  delete spectrogram_plot_;
    if (waterfall_data_)    delete waterfall_data_;
    if (raster_data_)       delete raster_data_;
}


void WaterfallPlot::init(int spectrogram_data_width, int spectrogram_data_depth, double max_frequency, double data_min, double data_max, int zoom_init, double span_init, double center_frequency_init) {
    spectrogram_data_width_ = spectrogram_data_width;
    spectrogram_data_depth_ = spectrogram_data_depth;

    max_frequency_ = max_frequency;
    frequency_span_ = span_init;
    zoom_level_ = zoom_init;
    center_frequency_ = center_frequency_init;
    start_frequency_ = center_frequency_ - frequency_span_/2;
    stop_frequency_ = center_frequency_ + frequency_span_/2;

    data_min_ = data_min;
    data_max_ = data_max;

    if (waterfall_data_) delete waterfall_data_;
    waterfall_data_ = new WaterfallData(spectrogram_data_width_, spectrogram_data_depth_);

    if (raster_data_) delete raster_data_;
    raster_data_ = new RasterData(spectrogram_data_width_, spectrogram_data_depth_);

    raster_data_ -> set_axis(start_frequency_, 0.0, stop_frequency_, spectrogram_data_depth_);
    raster_data_ -> set_data_range(data_min, data_max);
    raster_data_ -> set_data( waterfall_data_ -> get_ptr_to_array() );   // sets the data pointer of the RasterData class to the data array in the WaterfallData class



    // QwtPlotSpectrogram
    if (spectrogram_plot_) delete spectrogram_plot_;
    spectrogram_plot_ = new QwtPlotSpectrogram();

    // set a default color map
    QwtLinearColorMap waterfall_colors(Qt::blue, Qt::magenta);
    waterfall_colors.addColorStop(0.2, Qt::green);
    waterfall_colors.addColorStop(0.3, Qt::yellow);
    waterfall_colors.addColorStop(0.5, Qt::red);

    spectrogram_plot_ -> setColorMap(waterfall_colors);
    spectrogram_plot_ -> setData(*raster_data_);
    spectrogram_plot_ -> attach(this);
    spectrogram_plot_ -> setDisplayMode(QwtPlotSpectrogram::ImageMode, true);


    // default scale plot
    this -> setAxisScale(QwtPlot::xBottom, start_frequency_, stop_frequency_);
    this -> setAxisScale(QwtPlot::yLeft, 0, spectrogram_data_depth_);
    this -> enableAxis(QwtPlot::yLeft, false);  // do not display y axis

    // set scale plot fonts for X11
    QFont axisfont = this->axisFont(0); // read default font
    axisfont.setPointSize(8);
    this -> setAxisFont(QwtPlot::xBottom,axisfont);


    // default values for min and max zoom level
    max_zoom_ = 14;
    min_zoom_ = 0;

    //left_border_ = 44; // check this value, it seems that this should be sth like 7
    left_border_ = 7;
    right_border_ = left_border_ + spectrogram_data_width_;  // 1068 bei width 1024

    receiver_update_required_ = true;

    set_center_frequency(center_frequency_init);
}


void WaterfallPlot::correct_size() {

    cerr << "read pixel size of spectrogram plot.."<<endl;
    raster_data_ -> reset_pixel_counter();
    this -> replot();

    cerr << "current plot size:  x = " << raster_data_->get_pixels_x() << "px, y = " << raster_data_->get_pixels_y() << "px" << endl;
    cerr << "required plot size: x = " << spectrogram_data_width_ << "px, y = "<<spectrogram_data_depth_ << "px" << endl;

    if ((raster_data_->get_pixels_x() != spectrogram_data_width_) || (raster_data_->get_pixels_y() != spectrogram_data_depth_)) {

        cerr << "required to correct size.. ";

        //correct width
        int current_width = this -> width();
        int new_width = current_width + (spectrogram_data_width_ - raster_data_->get_pixels_x());

        //correct height
        int current_height = this -> height();
        int new_height = current_height + (spectrogram_data_depth_ - raster_data_->get_pixels_y());

        // set new size
        QRect plot_size_new(this -> geometry());
        plot_size_new.setWidth(new_width);
        plot_size_new.setHeight(new_height);
        this->setGeometry(plot_size_new);

        // read new size for checking
        raster_data_ -> reset_pixel_counter();
        this -> replot();
        cerr << "done" << endl;
        cerr << "     new plot size, x = " << raster_data_->get_pixels_x() << "px, y = " << raster_data_->get_pixels_y() << "px" << endl;
    }
    else
        cerr << "no action - size already correct" << endl;

}


void WaterfallPlot::set_next_data_line(double *array) {
    waterfall_data_ -> set_next_data_line(array);
    spectrogram_plot_->setData(*raster_data_);
    this -> replot();
}


void WaterfallPlot::set_center_frequency(double center_frequency) {
    // empty waterfall
    waterfall_data_ -> delete_all_data();

    // set new center_frequency, make sure, that span does not violate min and max frequency
    if ((center_frequency - frequency_span_/2) >= 0)
        center_frequency_ = center_frequency;
    else
        center_frequency_ = 0 + frequency_span_/2;

    if ((center_frequency + frequency_span_/2) <= max_frequency_)
        center_frequency_ = center_frequency;
    else
        center_frequency_ = max_frequency_- frequency_span_/2;

    // calculate start and stop frequency
    start_frequency_ = center_frequency_ - frequency_span_/2;
    stop_frequency_ = center_frequency_ + frequency_span_/2;


    // Scale Update
    this -> setAxisScale(QwtPlot::xBottom, start_frequency_, stop_frequency_);
    raster_data_ -> set_axis(start_frequency_, 0, stop_frequency_, spectrogram_data_depth_);
}



void WaterfallPlot::wheelEvent(QWheelEvent *event) {

    switch(wheel_action_) {
        case 0: // zoom with wheel
            // check, if mouse wheel turned in paint area
            if ((event->x() >= left_border_) && (event->x() <= right_border_)) {

                int zoom_position = event->x() - left_border_;

                // check, if mouse wheel was turned upwards and max zoom has not been reached
                if ( (event -> delta() > 0) && (zoom_level_ < max_zoom_) ) {
                    // zoom in
                    zoom_level_++;
                    cerr << "zoom in, now = " << zoom_level_ << endl;
                    zoom_update_axis( zoom_position, true );
                }

                // check, if mouse wheel was turned downwards and min zoom has not been reached
                if ( (event -> delta() < 0) && (zoom_level_ > min_zoom_) ) {
                    // zoom out
                    zoom_level_--;
                    cerr << "zoom out, now = " << zoom_level_ << endl;
                    zoom_update_axis( zoom_position , false );
                }

                receiver_update_required_ = true;
            }
            break;

        case 1: // shift frequency with wheel
            double minimum_possible_shift = frequency_span_/(double)spectrogram_data_width_;   // otherwise the plot gets wrongly shifted
            if (event -> delta() > 0) // wheel turned upwards
                shift_frequency(minimum_possible_shift);
            else
                shift_frequency(-minimum_possible_shift);

            break;
    }

}


void WaterfallPlot::zoom_update_axis(int x, bool zoom_in) {

    // determines frequency at zoom position
    double x_frequency = ((double)x / spectrogram_data_width_) * frequency_span_ + start_frequency_;

    // determines new frequency span
    //frequency_span_ = max_frequency_ / pow(2, zoom_level_);
    if (zoom_in)
        frequency_span_ = frequency_span_ / 2;
    else
        frequency_span_ = frequency_span_ * 2;

    // calculate start frequency
    if (zoom_in)
        start_frequency_ = x_frequency - (x_frequency - start_frequency_)/2 ; // for zoom in
    else
        start_frequency_ = x_frequency - (x_frequency - start_frequency_)*2;  // for zoom out


    // check if start frequency is below 0 or stop frequency is above max: then set new start/stop frequencies and shift the waterfall data to keep the frequencies aligned with the data
    // if new start frequency is below 0: shift data and set start frequency to 0
    if (start_frequency_ < 0) {
        int shift_idx = -2*(int)((start_frequency_/frequency_span_) * spectrogram_data_width_);
        waterfall_data_ -> shift_data_array(shift_idx);
        start_frequency_ = 0;
    }

    // if new stop frequency exceeds maximum value: shift data and set stop frequency to max and correct start frequency
    stop_frequency_ = start_frequency_ + frequency_span_;
    if (stop_frequency_ > max_frequency_) {
        int shift_idx = -2*(int)(((stop_frequency_-max_frequency_)/frequency_span_) * spectrogram_data_width_);
        waterfall_data_ -> shift_data_array(shift_idx);
        stop_frequency_ = max_frequency_;
        start_frequency_ = stop_frequency_ - frequency_span_;
    }

    // calculate new center frequency
    center_frequency_ = (stop_frequency_ + start_frequency_) / 2;

    // update zoom frequency (corrects x_frequency if start or stop has exceeded min/max value)
    x_frequency = ((double)x / spectrogram_data_width_) * frequency_span_ + start_frequency_;

    // sets the new scale of the displayed frequency axis of the QwtPlot
    this -> setAxisScale(QwtPlot::xBottom, start_frequency_, stop_frequency_);

    // sets the new scale for the raaster data class (takes care of proper assignment for the value (x,y) function)
    raster_data_ -> set_axis(start_frequency_, 0, stop_frequency_, spectrogram_data_depth_);

    // calculates the zoom position as index in the data array
    int zoom_center_idx = (int)(( (x_frequency-start_frequency_) / frequency_span_ ) * spectrogram_data_width_ +0.5);

    // zoom data array
    if (zoom_in)
        waterfall_data_ -> zoom_data_in(zoom_center_idx);
    else
        waterfall_data_ -> zoom_data_out(zoom_center_idx);

    // show zoomed spectrogram
    spectrogram_plot_ -> setData(*raster_data_);          // immediately changes the display according to the new scale
    this -> replot();

}


// shitfting of waterfall
void WaterfallPlot::mousePressEvent(QMouseEvent *event) {
    // save starting frequency for drag-and-drop frequency shift
    scroll_start_frequency_ = (((double) event->x()-left_border_) / spectrogram_data_width_) * frequency_span_ + start_frequency_; // with left_border
    //scroll_start_frequency_ = (((double) event->x()) / spectrogram_data_width_) * frequency_span_ + start_frequency_;

	// checks for right click = tune audio receiver
    if (event->button() == Qt::RightButton) emit this->audio_frequency_clicked(scroll_start_frequency_);
}

void WaterfallPlot::mouseMoveEvent(QMouseEvent *event) {

    double x_frequency = (((double)event->x()-left_border_) / spectrogram_data_width_) * frequency_span_ + start_frequency_; // with left_border
    //double x_frequency = (((double)event->x()) / spectrogram_data_width_) * frequency_span_ + start_frequency_;

    double frequency_shift = scroll_start_frequency_ - x_frequency;
    shift_frequency(frequency_shift);

}


void WaterfallPlot::shift_frequency(double frequency_shift) {
    double start_frequency_old = start_frequency_;
    double stop_frequency_old = stop_frequency_;

    start_frequency_ += frequency_shift;
    stop_frequency_ += frequency_shift;

    if (start_frequency_ < 0){
        start_frequency_ = 0;
        stop_frequency_  = frequency_span_;
        frequency_shift = -start_frequency_old;
    }

    if (stop_frequency_ > max_frequency_) {
        stop_frequency_ = max_frequency_;
        start_frequency_ = stop_frequency_ - frequency_span_;
        frequency_shift = max_frequency_ - stop_frequency_old;
    }

    center_frequency_ = (stop_frequency_ + start_frequency_) / 2;

    this -> setAxisScale(QwtPlot::xBottom, start_frequency_, stop_frequency_);      // scales the displayed axis of the QwtPlot
    raster_data_ -> set_axis(start_frequency_, 0, stop_frequency_, spectrogram_data_depth_);

    int shift_idx = (int)((frequency_shift/frequency_span_) * spectrogram_data_width_);

    waterfall_data_ -> shift_data_array(shift_idx);             // shift waterfall data

    spectrogram_plot_ -> setData(*raster_data_);
    replot();

    receiver_update_required_ = true;
}