#ifndef WATERFALLPLOT_H
#define WATERFALLPLOT_H

#include <QWheelEvent>

#include <qwt_plot.h>
#include <qwt_plot_spectrogram.h>
#include <qwt_color_map.h>

#include "waterfall/rasterdata.h"
#include "waterfall/waterfalldata.h"

/** @brief Top class for Waterfall Plots , 
*    - handles a complete waterfall display (Generic Widget on GUI is promoted to this class)
*    - includes data structres (RasterData & 2d array in WaterfallData) and plots (QwtPlot and QwtSpectrogram) */
class WaterfallPlot : public QwtPlot
{
    Q_OBJECT
public:
	/** @brief constructor */
    explicit WaterfallPlot(QWidget *parent = 0);

	/** @brief destructor */
    ~WaterfallPlot();

    /** @brief initializing function, creates data structures for spectrogram and its display
     ** important: the user needs to make sure, that the values specified here fit to the data, which is passed in set_next_data_line
     **            that means: zoom level and its corresponding span needs to fit the fft data passed to this class!
     **            if this is not ensured, the displayed waterfall has nothing to do with the frequency scale displayed
     *  @param spectrogram_data_width width of 2 dimensional data array (and display area in px)
     *  @param spectrogram_data_depth depth/height of 2 dimensional data array (and display area in px)
     *  @param max_frequency maximum frequency for display scale in kHz
     *  @param data_min expected minimum data value required by rasterdata.h
     *  @param data_max expected maximum data value required by rasterdata.h
     *  @param zoom_init inital zoom level
     *  @param span_init inital span in KHz
     *  @param center_frequency_init inital center frequency in kHz */
    void init(int spectrogram_data_width, int spectrogram_data_depth, \
              double max_frequency, double data_min, double data_max,  \
              int zoom_init, double span_init, double center_frequency_init);

    /** @brief sets the maximum possible zoom level */
    void set_max_zoom(int max_zoom) {max_zoom_ = max_zoom;}

    /** @brief sets the minimum possible zoom level */
    void set_min_zoom(int min_zoom) {min_zoom_ = min_zoom;}

    /** @brief changes the color map of the waterfall plot (a default is assinged in the constructor)   */
    void set_color_map(QwtLinearColorMap color_map) {spectrogram_plot_->setColorMap(color_map); this->replot();}

    /** @brief sets one line in the 2 dim data array for spectrogram
     *  @param array double array for one line  */
    void set_next_data_line(double *array);

    /** @brief corrects the size of the plot area, so that the data width always matches the display width in pixels */
    void correct_size();

    /** @brief sets center frequency in kHz (in this case the waterfall display is emptied completely) */
    void set_center_frequency(double center_frequency);

    /** @brief sets, what happens if the user turn the mouse wheel in the plot area
     *  @param wheel_action 0=zoom with wheel (default), 1=shift frequency by a small frequency (e.g. 20 Hz for audio waterfall) */
    void set_wheel_action(int wheel_action) { wheel_action_ = wheel_action; }

	/** @ brief returns if a receiver update is required (happens if a change in the waterfall has been made, e.g. zoom or shift)*/
    bool receiver_update_required() {return receiver_update_required_;}
	
	/** @brief confirms, that a receiver update has been recognized and resets the internal memeber, which signals "update required"*/
    void reset_receiver_update() {receiver_update_required_ = false;}

	// getter functions
	int    get_zoom_level() 		{return zoom_level_;}
    double get_center_frequency() 	{return center_frequency_;}
    double get_start_frequency()  	{return start_frequency_;}
    double get_stop_frequency()   	{return stop_frequency_;}
    double get_frequency_span()  	{return frequency_span_;}
    double get_max_frequency()  	{return max_frequency_;}
    double get_data_min()       	{return data_min_;}
    double get_data_max()       	{return data_max_;}

signals:
	/** @brief sends the frequency for tuning the audio receiver (when right clicking in the waterfall)
      * @param frequency frequency in kHz */
    void audio_frequency_clicked(double frequency);
    
private:

    /** @brief handels wheel events (for zoom) and decides if zoom in or out or nothing (max or min zoom exeeded) is done and calls method zoom_update_axis  */
    void wheelEvent(QWheelEvent *event); 

    /** @brief calls zoom functions of waterfalldata, calculates new frequencies (start/stop/span etc.) and sets new plot scales
      * @param x zoom position (position of mouse pointer relative to left border of plot)
      * @param zoom_in true: zoom in, false: zoom out   */
    void zoom_update_axis(int x, bool zoom_in);

	/** @brief shifts the waterfall ny the specified shift */
    void shift_frequency(double frequency_shift);

	/** @brief event handler for mouse moves for shift functionality */
    void mouseMoveEvent(QMouseEvent *event);
	
	/** @brief event handler for mouse clicks for shift functionality and audio tuning */
    void mousePressEvent(QMouseEvent *event);

    QwtPlotSpectrogram  	*spectrogram_plot_;	// spectromgram plot (waterfall) integrated in the QwtPlot
    RasterData          	*raster_data_;		// data for spectrogram plot
    WaterfallData       	*waterfall_data_;   // holds a 2D array

    int zoom_level_;
    int max_zoom_;
    int min_zoom_;
    double data_min_;
    double data_max_;
    double max_frequency_;
    double start_frequency_;    // start frequency of waterfall display in kHz
    double stop_frequency_;     // stop frequency of waterfall display in kHz
    double center_frequency_;   // center frequency of waterfall display in kHz
    double frequency_span_;     // frequency span in waterfall display in kHz

    int spectrogram_data_width_; // spectrogram width in pixels/data points
    int spectrogram_data_depth_; // spectrogram height/depth in pixels/data points

    double scroll_start_frequency_;	// stores the start frequency at the beginning of a shift
    int left_border_;            // marks left border of waterfall painting area
    int right_border_;           // marks right border of waterfall painting area

    bool receiver_update_required_; // signals, that a scroll or zoom activity has occured and the receiver has to be updated

    int wheel_action_;			// specifies, what happend if the wheel is turned (see set_wheel_action())
};

#endif // WATERFALLPLOT_H