#include "waterfallddc.h"

WaterfallDDC::WaterfallDDC()
{
    fft_real_ = NULL;
    fft_complex_ = NULL;
    local_oscillator_ = NULL;
    cic_filter_ = NULL;
    cic_bypass_ = NULL;
    iq_capture_module_ = NULL;
    rf_capture_module_ = NULL;

    spectrum_line_curve_ = NULL;
    spectrum_line_grid_ = NULL;

    captured_data_rf_ = NULL;
    captured_data_i_ = NULL;
    captured_data_q_ = NULL;
    spectrum_rf_ = NULL;
    spectrum_iq_ = NULL;
    spectrum_temp_ = NULL;
    spectrum_data_ = NULL;
    spectrum_display_ = NULL;

    lo_initialized_ = false;
    cic_initialized_ = false;
    capture_iq_initialized_ = false;
    capture_rf_initialized_ = false;
    waterfall_initialized_ = false;
    labels_initialized_ = false;
    fft_initialized_ = false;
    line_plot_initialized_ = false;

    debug_enabled_ = false;

    capture_iq_max_size_ = -1;
    capture_rf_max_size_ = -1;
    fft_rf_size_ = -1;
    fft_iq_size_ = -1;
    data_size_ = -1;

    num_ffts_ = 1;
}

WaterfallDDC::~WaterfallDDC()
{
    if (fft_real_) delete fft_real_;
    if (fft_complex_) delete fft_complex_;
    if (local_oscillator_) delete local_oscillator_;
    if (cic_filter_) delete cic_filter_;
    if (cic_bypass_) delete cic_bypass_;
    if (iq_capture_module_) delete iq_capture_module_;
    if (rf_capture_module_) delete rf_capture_module_;

    if(spectrum_line_curve_) delete spectrum_line_curve_;
    if(spectrum_line_grid_) delete spectrum_line_grid_;

    if (captured_data_rf_) delete[] captured_data_rf_;
    if (captured_data_i_) delete[] captured_data_i_;
    if (captured_data_q_) delete[] captured_data_q_;
    if (spectrum_rf_) delete[] spectrum_rf_;
    if (spectrum_iq_) delete[] spectrum_iq_;
    if (spectrum_data_) delete[] spectrum_data_;
    if (spectrum_display_) delete[] spectrum_display_;
    if (spectrum_temp_) delete[] spectrum_temp_;

}

void WaterfallDDC::init_lo(int address, int memsize, double master_frequency, int phase_width, double init_frequency) {
    if (local_oscillator_) delete local_oscillator_;
    local_oscillator_ = new DDS(address, memsize, master_frequency, phase_width);
    local_oscillator_ -> set_frequency(init_frequency);
    lo_initialized_ = true;
}

void WaterfallDDC::init_cic(int filter_address, int filter_memsize, int bypass_address, int bypass_memsize) {
    if (cic_filter_) delete cic_filter_;
    cic_filter_ = new CIC(filter_address, filter_memsize, 8);

    if (cic_bypass_) delete cic_bypass_;
    cic_bypass_ = new IP_Driver(bypass_address, bypass_memsize);

    cic_initialized_ = true;
    set_cic(0);
}

void WaterfallDDC::init_iq_capture(int address, int memsize, int capture_size) {
    if (iq_capture_module_) delete iq_capture_module_;
    iq_capture_module_ = new Data_Capture_Dual(address, memsize, capture_size);
    capture_iq_max_size_ = capture_size;
    capture_iq_initialized_ = true;
}

void WaterfallDDC::init_rf_capture(int address, int memsize, int capture_size) {
    if (rf_capture_module_) delete rf_capture_module_;
    rf_capture_module_ = new Data_Capture(address, memsize, capture_size);
    capture_rf_max_size_ = capture_size;
    capture_rf_initialized_ = true;
}

void WaterfallDDC::init_waterfall(WaterfallPlot *waterfall, double max_frequency, double data_range_min, double data_range_max, double span_init, int data_size, int waterfall_depth, double usable_span_frac) {

    if ((lo_initialized_) && (cic_initialized_) && (capture_iq_initialized_) && (capture_rf_initialized_) && (capture_rf_max_size_ >= data_size*2) && (capture_iq_max_size_ >= data_size*2)) {
        data_size_ = data_size;
        display_size_ = (int)floor(usable_span_frac * data_size_);
        if (display_size_ % 2 == 1) display_size_++;                // ensure, that display_size is an equal number as required by waterfall class

        double usable_span_frac_true = (double)display_size_ / (double)data_size_; // determine actual usable span frac
        cerr << "requested span frac = " << usable_span_frac << ", chosen span frac = " << usable_span_frac_true << ", data_width = " << data_size_ << ", display width = " << display_size_ << endl;

        waterfall_ = waterfall;
        waterfall_ -> init(display_size_, waterfall_depth, max_frequency*usable_span_frac_true, data_range_min, data_range_max, 0, span_init*usable_span_frac_true, (span_init*usable_span_frac_true)/2);
        waterfall_ -> set_min_zoom(0);
        waterfall_ -> set_max_zoom(14);

        connect(waterfall_, SIGNAL(audio_frequency_clicked(double)),this, SIGNAL(audio_frequency_clicked(double)));

        waterfall_initialized_ = true;
    }
    else {
        cerr << "ERROR waterfallddc.cpp: run init_lo, init_cic, init_iq_capture and init_rf_capture before init_waterfall or check fft/waterfall data size with max data size of capture IP core " << endl;
        exit(1);
    }
}


void WaterfallDDC::init_labels(QLabel* label_zoom, QLabel* label_resolution, QLabel* label_bandwidth, QLabel* label_lo_freq, QLabel* label_cic_dec, QLabel* label_fir_dec, QLabel* label_fft) {
    label_zoom_ = label_zoom;
    label_resolution_ = label_resolution;
    label_bandwidth_ = label_bandwidth;
    label_lo_freq_ = label_lo_freq;
    label_cic_dec_ = label_cic_dec;
    label_fir_dec_ = label_fir_dec;
    label_fft_ = label_fft;
    labels_initialized_ = true;
}



void WaterfallDDC::init_fft(int fft_rf_size, int fft_iq_size, int window_type) {
    if ((!capture_iq_initialized_)|| (!capture_rf_initialized_) || (!waterfall_initialized_) ) {
        cerr << "ERROR: waterfallddc.cpp init_fft: run init_capture, init_waterfall and init_rf_capture first" << endl;
        exit(1);
    }

    if ((fft_rf_size < data_size_*2) || (fft_rf_size > capture_rf_max_size_) \
      ||(fft_iq_size < data_size_*2) || (fft_iq_size > capture_iq_max_size_)) {
        cerr << "ERROR: waterfallddc.cpp init_fft: invalid fft size specified" << endl;
        exit(1);
    }

    fft_rf_size_ = fft_rf_size;
    fft_iq_size_ = fft_iq_size;

    if (fft_real_) delete fft_real_;
    if (fft_complex_) delete fft_complex_;

    switch(window_type) {
        case 0: fft_real_    = new fft_avg_real(fft_rf_size_, fftw_base::no);
                fft_complex_ = new fft_avg_complex(fft_iq_size_, fftw_base::no);
                break;

        case 1: fft_real_    = new fft_avg_real(fft_rf_size_, fftw_base::hann);
                fft_complex_ = new fft_avg_complex(fft_iq_size_, fftw_base::hann);
                break;

        case 2: fft_real_    = new fft_avg_real(fft_rf_size_, fftw_base::flat);
                fft_complex_ = new fft_avg_complex(fft_iq_size_, fftw_base::flat);
                break;

        default: cerr << "ERROR waterfallddc.cpp: init_fft: invalid window type specified" << endl;
                 exit(1);
    }


    fft_initialized_ = true;


    if (captured_data_rf_) delete[] captured_data_rf_;
    captured_data_rf_ = new double[fft_rf_size_];

    if (captured_data_i_) delete[] captured_data_i_;
    captured_data_i_ = new double[fft_iq_size_];

    if (captured_data_q_) delete[] captured_data_q_;
    captured_data_q_ = new double[fft_iq_size_];

    if (spectrum_rf_) delete[] spectrum_rf_;
    spectrum_rf_ = new double[fft_rf_size_/2];

    if (spectrum_iq_) delete[] spectrum_iq_;
    spectrum_iq_ = new double[fft_iq_size_];

    if (spectrum_data_) delete[] spectrum_data_;
    spectrum_data_ = new double[data_size_];

    if (spectrum_display_) delete[] spectrum_display_;
    spectrum_display_ = new double[display_size_];

    if (spectrum_temp_) delete[] spectrum_temp_;
    spectrum_temp_ = new double[data_size_*2];

}


void WaterfallDDC::init_spectrum_line_plot(QwtPlot *spectrum_line_plot) {
    if ((!capture_iq_initialized_)|| (!capture_rf_initialized_) || (!waterfall_initialized_) ) {
        cerr << "ERROR: waterfallddc.cpp init_spectrum_line_plot: run init_iq_capture, init_waterfall and init_rf_capture first" << endl;
        exit(1);
    }

    // set pointer to QwtPlot specified by caller and set Axes and Fonts
    spectrum_line_plot_ = spectrum_line_plot;

    QFont axisfont = spectrum_line_plot_->axisFont(0); // read default font
    axisfont.setPointSize(8);
    spectrum_line_plot_ -> setAxisFont(0,axisfont);
    spectrum_line_plot_ -> setAxisFont(2,axisfont);
    spectrum_line_plot_ -> setAxisScale(0,waterfall_->get_data_min(), waterfall_->get_data_max(), 10);    // y axis, axis step: 10
    spectrum_line_plot_ -> setAxisScale(2, waterfall_->get_start_frequency(), waterfall_->get_stop_frequency(), 0);   // x axis

    QwtText ytitle("Power [dBm]");
    axisfont.setPointSize(10);
    axisfont.setBold(true);
    ytitle.setFont(axisfont);
    spectrum_line_plot_ -> setAxisTitle(QwtPlot::yLeft, ytitle);

    QwtText xtitle("Frequency [kHz]");
    axisfont.setPointSize(10);
    axisfont.setBold(true);
    xtitle.setFont(axisfont);
    spectrum_line_plot_ -> setAxisTitle(QwtPlot::xBottom, xtitle);


    // define a plot curve and attach it to the QwtPlot
    spectrum_line_curve_ = new QwtPlotCurve();
    spectrum_line_curve_ -> setPen(QPen(Qt::blue, 2));
    spectrum_line_curve_ -> attach(spectrum_line_plot_);

    // attach a grid to the QwtPlot
    spectrum_line_grid_ = new QwtPlotGrid;
    spectrum_line_grid_ -> attach(spectrum_line_plot_);

    line_plot_initialized_ = true;
}


void WaterfallDDC::set_color_map(QwtLinearColorMap color_map) {
    if (!waterfall_initialized_) {
        cerr << "ERROR: waterfallddc.cpp set_color_map: run init_waterfall first" << endl;
        exit(1);
    }

    waterfall_->set_color_map(color_map);
}



void WaterfallDDC::capture_and_show(fstream &record_file, bool record_data) {
    if (!waterfall_initialized_) {
        cerr << "ERROR: waterfall_ddc.cpp: run init_waterfall before calling capture_and_show" << endl;
        exit(1);
    }

    // suppress function execution if debug mode is enabled
    if (debug_enabled_) return;

    QTime zeitmessung;
    zeitmessung.start();
    int receiver_update_time = 0;
    int capture_fft_time = 0;
    int display_spectrum_time = 0;
    int display_waterfall_time = 0;

    // 1. update receiver parameter ( this part was at the end (3.) in the former version, if I remember correctly due to some display bugs when zooming or shifting. If problem occur again, try working on this part of the code)
    set_lo_frequency(waterfall_ -> get_center_frequency());

    switch (waterfall_ -> get_zoom_level()) {
        case 0:     set_cic(0); break;
        case 1:     set_cic(0); break;
        case 2:     set_cic(0); break;
        case 3:     set_cic(0); break;
        case 4:     set_cic(4); break;
        case 5:     set_cic(8); break;
        case 6:     set_cic(16); break;
        case 7:     set_cic(32); break;
        case 8:     set_cic(64); break;
        case 9:     set_cic(128); break;
        case 10:    set_cic(256); break;
        case 11:    set_cic(512); break;
        case 12:    set_cic(1024); break;
        case 13:    set_cic(2048); break;
        case 14:    set_cic(4096); break;
        default:    cerr << "zoom level not supported" << endl;
    }


    if (waterfall_ -> receiver_update_required()) {
        if (labels_initialized_) {
            // zoom, resolution and bandwidth label
            label_zoom_ -> setText(QString::number(waterfall_->get_zoom_level())+ " x");                           // zoom level
            label_resolution_ -> setText(QString::number(waterfall_->get_frequency_span()/data_size_,'f',3) + " kHz");   // resolution
            label_bandwidth_ -> setText(QString::number(waterfall_->get_frequency_span(),'f',0)  + " kHz");        // bandwidth

            // FIR dec and FFT label
            if ((waterfall_ -> get_zoom_level() == 0) || (waterfall_ -> get_zoom_level() == 1)) {
                label_fft_ -> setText(QString::number(fft_rf_size_) + " real");
                label_fir_dec_ -> setText("-");
            }
            else if ((waterfall_ -> get_zoom_level() >=2 ) && (waterfall_ -> get_zoom_level() <= 14)) {
                label_fft_ -> setText(QString::number(fft_iq_size_) + " IQ");
                label_fir_dec_ -> setText("8");
            }
            else {
                label_fft_ -> setText("n/a");
                label_fir_dec_ -> setText("8");
            }

        }

        // reset update request flags*/
        waterfall_ -> reset_receiver_update();

    }

    receiver_update_time = zeitmessung.elapsed();


    // 2. capture data according to zoom_level

    int start_idx;

    switch (waterfall_ -> get_zoom_level()) {

        // zoomlevel 0: full span:
        case 0:
            // capture rf data and calculate averaged FFT
            fft_real_ -> reset();

            for (int fft_idx = 0; fft_idx < num_ffts_; fft_idx++) {
                rf_capture_module_ -> capture_and_read(captured_data_rf_, fft_rf_size_);
                fft_real_ -> add_fft(captured_data_rf_);
            }

            fft_real_ -> get_avg_fft(spectrum_rf_);

            // correct gain of calculated spectrum
            for (int i = 0; i < (fft_rf_size_/2); i++)
                spectrum_rf_[i] = spectrum_rf_[i] - gain_rf_scaler_;

            // decimate the spectrum, such that it fits into the display
            decimate_spectrum(spectrum_rf_, fft_rf_size_/2, spectrum_data_, data_size_);

            // cut out leftmost bins for display (omit frequencies with aliasing)
            for (int i = 0; i < display_size_; i++)
                spectrum_display_[i] = spectrum_data_[i];

            break;


        // zoom_level 1:half span:
        case 1:
            // capture rf data and calculate averaged FFT
            fft_real_ -> reset();

            for (int fft_idx = 0; fft_idx < num_ffts_; fft_idx++) {
                rf_capture_module_ -> capture_and_read(captured_data_rf_, fft_rf_size_);
                fft_real_ -> add_fft(captured_data_rf_);
            }

            fft_real_ -> get_avg_fft(spectrum_rf_);

            // correct gain of calculated spectrum
            for (int i = 0; i < (fft_rf_size_/2); i++)
                spectrum_rf_[i] = spectrum_rf_[i] - gain_rf_scaler_;

            // decimate the spectrum, such that it fits into the display
            decimate_spectrum(spectrum_rf_, fft_rf_size_/2, spectrum_temp_, data_size_*2);

            // select data for display
            start_idx = (display_size_*2) * (waterfall_->get_start_frequency() / waterfall_->get_max_frequency());

            // cut out bins for display
            for (int i=0; i < data_size_ ;i++)
                spectrum_data_[i] = spectrum_temp_[(int)start_idx + i];

            // cut out left most bins for display to omit aliasing
            for (int i = 0; i < display_size_; i++)
                spectrum_display_[i] = spectrum_data_[i];

            break;


        // zoomlevel 3: complex FFT results, cut out bins for display
        case 3:

            // capture iq data and calculate averaged FFT
            fft_complex_->reset();

            for (int fft_idx = 0; fft_idx < num_ffts_; fft_idx++) {
                iq_capture_module_ -> capture_and_read(captured_data_i_, captured_data_q_, fft_iq_size_);
                fft_complex_ -> add_fft(captured_data_i_, captured_data_q_);
            }

            fft_complex_ -> get_avg_fft(spectrum_iq_);

            // correct gain of calculated spectrum
            for (int i = 0; i < fft_iq_size_; i++)
                spectrum_iq_[i] = spectrum_iq_[i] - gain_ddc_ -gain_rf_scaler_;

            // decimate the spectrum, such that it fits into the display
            decimate_spectrum(spectrum_iq_, fft_iq_size_, spectrum_temp_, data_size_*2);


            // select data for display
            start_idx = (data_size_*2) * (waterfall_->get_start_frequency() / waterfall_->get_max_frequency());
            start_idx = (data_size_*2)/2 - data_size_/2; // start chosen such that the spectrum with data_size_ bins lies exactly in the center of the data_size_*2 bins

            // cut out bins for display
            for (int i=0; i < data_size_ ;i++)
                spectrum_data_[i] = spectrum_temp_[(int)start_idx + i];

            // cut out center bins for display to omit aliasing
            for (int i = 0; i < display_size_; i++)
                spectrum_display_[i] = spectrum_data_[i + ( (data_size_-display_size_) / 2 ) ];

            break;

        case 2:
        case 4:
        case 5:
        case 6:
        case 7:
        case 8:
        case 9:
        case 10:
        case 11:
        case 12:
        case 13:
        case 14:

            // capture iq data and calculate averaged FFT
            fft_complex_->reset();

            for (int fft_idx = 0; fft_idx < num_ffts_; fft_idx++) {
                iq_capture_module_ -> capture_and_read(captured_data_i_, captured_data_q_, fft_iq_size_);
                fft_complex_ -> add_fft(captured_data_i_, captured_data_q_);
            }

            fft_complex_ -> get_avg_fft(spectrum_iq_);

            // correct gain of calculated spectrum
            for (int i = 0; i < fft_iq_size_; i++)
                spectrum_iq_[i] = spectrum_iq_[i] - gain_ddc_ - gain_rf_scaler_;

            // decimate the spectrum, such that it fits into the display
            decimate_spectrum(spectrum_iq_, fft_iq_size_, spectrum_data_, data_size_);

            // cut out center bins for display to omit aliasing
            for (int i = 0; i < display_size_; i++)
                spectrum_display_[i] = spectrum_data_[i + ( (data_size_-display_size_) / 2 ) ];

            break;


        default:
            cerr << "not yet implemented " << endl;

    }

    capture_fft_time = zeitmessung.elapsed();


    // 3. display generated data line in waterfall
    if (waterfall_->receiver_update_required() == false)
        waterfall_ -> set_next_data_line(spectrum_display_);            // Waterfall display (if waterfall diagram has not been modified = receiver update required)

    display_waterfall_time = zeitmessung.elapsed();


    // 4. display spectrum in line plot
    if (line_plot_initialized_) {
        double frequency_scale[display_size_];
        for (int i=0; i<display_size_; i++)
            frequency_scale[i] = waterfall_->get_start_frequency() + i * (waterfall_->get_frequency_span() / (double)(display_size_));     // Spectrum Line Display

        spectrum_line_plot_ -> setAxisScale(2, waterfall_->get_start_frequency(), waterfall_->get_stop_frequency(), 0);
        spectrum_line_curve_ -> setRawData(frequency_scale, spectrum_display_, display_size_);
        spectrum_line_plot_ -> replot();
    }

    display_spectrum_time = zeitmessung.elapsed();

    if (record_data) {
        for (int i=0; i<display_size_; i++)
            record_file << spectrum_display_[i] << ", ";

        record_file << endl;
    }

    cerr << "waterfall times: " << \
            "update: " << receiver_update_time << ", " << \
            "cap & fft: " << capture_fft_time - receiver_update_time << ", " << \
            "waterfall: "<< display_waterfall_time - capture_fft_time << ", " << \
            "spectrum: " << display_spectrum_time - display_waterfall_time << endl;

}


// ---------------------------------------------------------------------------------------------------------------------------------------------------------------
// private methods


void WaterfallDDC::set_cic(int decimation_rate) {
    if (!cic_initialized_) {
        cerr << "ERROR: cic not initialized" << endl;
        exit(1);
    }

    if (decimation_rate <=1) {
        cic_bypass_ -> write(0,0);                        // set cic bypass
        if (labels_initialized_)
            label_cic_dec_ -> setText("bypassed");
    }
    else {

        if (decimation_rate <= 512) {
            cic_bypass_ -> write(0,0xFFFFFFF1);                             // select programmable rate cic and bypass fixed rate cic
            cic_filter_ -> set_rate(decimation_rate);                       // set CIC rate
        }
        else {
            cic_bypass_ -> write(0,0xFFFFFFF3);                             // select programmable rate cic and fixed rate cic
            cic_filter_ -> set_rate(decimation_rate/64);                    // set CIC rate
        }

        if (labels_initialized_)
            label_cic_dec_ -> setText(QString::number(decimation_rate));     // display total CIC decimation rate

    }
}


double WaterfallDDC::set_lo_frequency(double lo_frequency) {
    if (!lo_initialized_) {
        cerr << "ERROR: audio_ddc.cpp: LO not initialized, call init_lo before setting the RX frequency" << endl;
        exit(1);
    }

    double true_freq = local_oscillator_ -> set_frequency(lo_frequency);

    if (labels_initialized_)
        label_lo_freq_ -> setText(QString::number(true_freq,'f',0) + "kHz");

    return true_freq;
}



void WaterfallDDC::decimate_spectrum(double data_in[], int data_in_size, double data_out[], int data_out_size) {
    if (data_in_size % data_out_size != 0) {
        cerr << "ERROR WaterfallDDC: decimate_spectrum: data_in_size ("<< data_in_size << ") and/or data_out_size (" << data_out_size << ") do not fit" << endl;
        exit(1);
    }

    int decimation_factor = data_in_size / data_out_size;

    double max_temp = -99999;

    for (int i = 0; i < data_out_size; i++) {
        max_temp = -99999;

        for (int j= i* decimation_factor; j < (i+1)*decimation_factor ;j++)
            if (data_in[j] > max_temp)
                max_temp = data_in[j];

        data_out[i] = max_temp;
    }

}


// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//   Debug Functions
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void WaterfallDDC::set_debug(bool debug_en) {
     debug_enabled_ = debug_en;

     cerr << endl << "Debug Mode enabled " << endl;
}

void WaterfallDDC::debug_set_cic(int dec) {
    if (debug_enabled_)
        set_cic(dec);
    else
        cerr << "CIC cannot be set, enable debug mode first" << endl;
}

void WaterfallDDC::debug_set_lo(double freq) {
    if (debug_enabled_)
        set_lo_frequency(freq);
    else
        cerr << "LO frequency cannot be set, enable debug mode first" << endl;
}


void WaterfallDDC::debug_capture_iq_samples(double samples_i[], double samples_q[], int size) {
    if (debug_enabled_)
        iq_capture_module_ -> capture_and_read(samples_i, samples_q, size);
    else
        cerr << "IQ data cannot be captured, enable debug mode first" << endl;
}

void WaterfallDDC::debug_capture_rf_samples(double samples[], int size) {
    if (debug_enabled_)
        rf_capture_module_ -> capture_and_read(samples, size);
    else
        cerr << "IQ data cannot be captured, enable debug mode first" << endl;
}
