#ifndef WATERFALLDDC_H
#define WATERFALLDDC_H

#include "../waterfall/waterfallplot.h"
#include <fft/fft_avg_complex.h>
#include <fft/fft_avg_real.h>
#include "stipcores.h"
#include <qwt_plot_curve.h>
#include <qwt_plot_grid.h>
#include <qlabel.h>

/** @brief Class for controlling the Waterfall DDC \n
 *  - first call init functions: first: init_lo, init_crc, init_iq_capture, second: init_waterfall, third: init_fft, init_spectrum_line_plot and optionally init_labels (in this order) to set up the class before use
 *  - use other functions to display a new line in the waterfall or to set other properties */
class WaterfallDDC : public QObject
{
    Q_OBJECT
public:

	/** @brief constructor */
    WaterfallDDC();
	
	/** @brief destructor */
    ~WaterfallDDC();


    /** @brief initializes LO IP core, required to run before using the DDC!
     *  @param address IP core address (from Vivado)
     *  @param memsize IP core memory size in k  (from Vivado)
     *  @param master_frequency DDS master frequency in kHz (from Vivado)
     *  @param phase_width DDS phase width (from Vivado)
     *  @param init_frequency inital LO frequency in kHz */
    void init_lo(int address, int memsize, double master_frequency, int phase_width, double init_frequency);

    /** @brief initializes CIC IP cores (rate prog and bypass), required to run before using the DDC!
     *  @param filter_address IP core address for programming interface IP core (from Vivado)
     *  @param filter_memsize IP core memory size in k for programming interface IP core  (from Vivado)
     *  @param bypass_address IP core address for CIC bypass interface IP core (from Vivado)
     *  @param bypass_memsize IP core memory size in k for CIC bypass interface IP core  (from Vivado) */
    void init_cic(int filter_address, int filter_memsize, int bypass_address, int bypass_memsize);

    /** @brief initializes Dual Capture IP core for IQ data, required to run before using the DDC!
     *  @param address IP core address (from Vivado)
     *  @param memsize IP core memory size in k  (from Vivado)
     *  @param capture_size depth of capture memory in i/q sample pairs */
    void init_iq_capture(int address, int memsize, int capture_size);

    /** @brief initializes the RF capture module, instanciated outside of this class
     *  @param address IP core address (from Vivado)
     *  @param memsize IP core memory size in k  (from Vivado)
     *  @param capture_size depth of capture memory samples */
    void init_rf_capture(int address, int memsize, int capture_size);

    /** @brief initializes the Waterfall DDC, required to run before use!
      * @param *waterfall pointer to waterfall display
      * @param max_frequency maximum frequency the waterfall can be scrolled to in kHz, give value for usable_span_frac = 1.0 here!!!
      * @param data_range_min minimum data value in waterfall
      * @param data_range_max maximum data value in waterfall
      * @param span_init initial frequency span of waterfall in kHz (= audio IQ signal's sample rate)
      * @param data_size size of data block for displayed (including aliasing bin, that will be truncated for final display)
      * @param waterfall_depth depth of waterfall display
      * @param usable_span_frac provides a fraction (0 to 1) of how much of the calculated spectrum can be used without aliasing effects*/
    void init_waterfall(WaterfallPlot *waterfall, double max_frequency, double data_range_min, double data_range_max, double span_init, int data_size, int waterfall_depth, double usable_span_frac);

    /** @brief initializes the FFT cores and data arrays for FFT and related processing
     *  @param fft_rf_size (input) size for real FFT processing of RF samples
     *  @param fft_iq_size (input and output) size for complex FFT processing of IQ samples
     *  @param window_type 0=no/rect, 1=hanning, 2=flattop window */
    void init_fft(int fft_rf_size, int fft_iq_size, int window_type);

    /** @brief OPTIONAL: initializes labels for status display in the GUI, in which the WaterfallDDC can write status information */
    void init_labels(QLabel *label_zoom, \
                     QLabel *label_resolution, \
                     QLabel *label_bandwidth, \
                     QLabel *label_lo_freq, \
                     QLabel *label_cic_dec, \
                     QLabel *label_fir_dec, \
                     QLabel *label_fft);


    /** @brief OPTIONAL: initializes support to display the spectrum in an extra line plot (QwtPlot)
      * @param spectrum_line_plot pointer to QwtPlot Object, that will display the line spectrum */
    void init_spectrum_line_plot(QwtPlot *spectrum_line_plot);


    /** @brief sets the color map of the plot (a default is assinged in the constructor)   */
    void set_color_map(QwtLinearColorMap color_map);

    /** @brief sets the DDC gain in dB, that will be subtracted from the FFT power bins (without RF scaler gain) */
    void set_gain_ddc(int gain_ddc) {gain_ddc_ = gain_ddc;}

    /** @brief sets the RF Scaler gain in dB, that will be subtracted from the FFT power bins */
    void set_gain_rf_scaler(int gain_rf_scaler) {gain_rf_scaler_ = gain_rf_scaler;}

    /** @brief sets number of FFTs, that are averaged */
    void set_num_ffts(int num_ffts) {num_ffts_ = num_ffts;}


    /** @brief captures IQ data (after DDC) and shows them as a new line in the waterfall plot  */
    void capture_and_show(fstream &record_file, bool record_data);

    /** @brief required to call to resize the plot to the data length */
    void correct_waterfall_size() { if (waterfall_initialized_)  waterfall_ -> correct_size(); }

	/** @brief get data block length for real fft (fft_rf_size_/2 output bins) */
    int get_fft_rf_size()   {return fft_rf_size_;}
	
	/** @brief get data block length for complex fft */
    int get_fft_iq_size()   {return fft_iq_size_;}
	
	/** @brief get data block length for complex fft */
    int data_size()         {return data_size_;}



    // +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    // Debug Functions


    /** @brief enables the debug mode, which stops the waterfall and allows to set the DDC parameters and IP cores directly (for debug) */
    void set_debug(bool debug_en);

    /** @brief forces the CIC filter to the given decimation factor (for debug) */
    void debug_set_cic(int dec);

    /** @brief forces the LO to the given frequency (for debug) */
    void debug_set_lo(double freq);

    /** @brief returns IQ captured data (for debug) */
    void debug_capture_iq_samples(double samples_i[], double samples_q[], int size);

    /** @brief returns RF captured data (for debug)*/
    void debug_capture_rf_samples(double samples[], int size);

signals:
	/** @brief sends the frequency for tuning the audio receiver (when right clicking in the waterfall)
      * @param frequency frequency in kHz */
    void audio_frequency_clicked(double frequency);


private:
    fft_avg_real        *fft_real_;
    fft_avg_complex     *fft_complex_;
    DDS                 *local_oscillator_;
    CIC                 *cic_filter_;
    IP_Driver           *cic_bypass_;
    Data_Capture_Dual   *iq_capture_module_;
    WaterfallPlot       *waterfall_;
    Data_Capture        *rf_capture_module_;

    // objects for line spectrum display
    QwtPlot             *spectrum_line_plot_;
    QwtPlotCurve        *spectrum_line_curve_;
    QwtPlotGrid         *spectrum_line_grid_;

    // QLabel objects for status outputs (optional feature)
    QLabel              *label_zoom_;
    QLabel              *label_resolution_;
    QLabel              *label_bandwidth_;
    QLabel              *label_lo_freq_;
    QLabel              *label_cic_dec_;
    QLabel              *label_fir_dec_;
    QLabel              *label_fft_;

    // arrays, that hold captured data, fft data and display data
    double              *captured_data_rf_;
    double              *captured_data_i_;
    double              *captured_data_q_;
    double              *spectrum_rf_;              // spectrum calculated in full resolution of RF signal
    double              *spectrum_iq_;              // spectrum calculated in full resolution of IQ signals
    double              *spectrum_temp_;            // decimated spectrum, that needs to be cut out for display (size: 2*data_size_)
    double              *spectrum_data_;            // spectrum, that will be displyed in the waterfall
    double              *spectrum_display_;

    // different block sizes specified for the waterfall ddc
    int                 capture_iq_max_size_;      // depth of capture ip core memory
    int                 capture_rf_max_size_;
    int                 fft_rf_size_;              // data block length for real fft (fft_rf_size_/2 output bins)
    int                 fft_iq_size_;              // data block length for complex fft
    int                 data_size_;                // data block length for display (including the aliasing bins)
    int                 display_size_;             // size of displayed data block (truncated due to aliasing)

    // variable, that show which init function has been activated
    bool                lo_initialized_;
    bool                cic_initialized_;
    bool                capture_iq_initialized_;
    bool                capture_rf_initialized_;
    bool                waterfall_initialized_;
    bool                fft_initialized_;
    bool                labels_initialized_;        // indiciate if optional feature has been activated
    bool                line_plot_initialized_;     // indiciate if optional feature has been activated

    bool                debug_enabled_;

    double              gain_ddc_;        // power gain of DDC in dB, that will be subtracted from FFT power bins (without gain of RF scaler)
    double              gain_rf_scaler_;  // additional power of RF Scaler in dB, that will be subtracted from FFT power bins

    int                 num_ffts_;      // number of FFTs for averaging

    void set_cic(int decimation_rate);
    double set_lo_frequency(double lo_frequency);
	
	/** @brief converts long spectrum in a shorter one, using simple max operation */
    void decimate_spectrum(double data_in[], int data_in_size, double data_out[], int data_out_size);

};

#endif // WATERFALLDDC_H
