/*
  -------------------------------------------------------
    AD9467_Data_Interface.cpp
    C++ driver class for ADC data interface (IO modules)
    Stefan Scholl, DC9ST, TU Kaiserslautern, 2014
  -------------------------------------------------------
*/


#include <iostream>
#include "AD9467_Data_Interface.h"

using namespace std;


// private functions

//wait for approx. t_ms ms */
void AD9467_Data_Interface::wait_ms(int t_ms) {

	volatile int a = 0;

	for (int j=0; j < t_ms; j++)
		for (int i=0; i<50000; i++)
			a += 1;
};


// constructor and destructor

// class constructor
AD9467_Data_Interface::AD9467_Data_Interface(int base_address, int size_in_k) {
	data_iface_core_ = new IP_Driver(base_address, size_in_k);
}

// class deconstructor	
AD9467_Data_Interface::~AD9467_Data_Interface() {
	delete data_iface_core_;
}


//public functions

// sets delay value for one delay element
int AD9467_Data_Interface::set_delay(int line_addr, int delay_value) {

	// check requested delay values for consistency
	if ( (delay_value > 31) || (delay_value < 0) ) return -2;
	if ( (line_addr > 8) || (line_addr < 0) ) return -3;

	// apply address of line
	data_iface_core_->write(0x00, line_addr);

	// apply we and delay data
	data_iface_core_->write(0x04, 1<<5 | delay_value); // we and delay value

	// wait for ack
	int delay_value_read;
	int timeout_counter = 0;

	do {
		delay_value_read = data_iface_core_->read(0x08);
		wait_ms(1);
		timeout_counter++;
	}
	while ( (delay_value_read != delay_value) && (timeout_counter < 10) );

	// cancel we and delay data
	data_iface_core_->write(0x04,0x00);  // we and delay value

	if (timeout_counter >= 10) return -1; 	// catch timeout

	return 0;
}


// @brief captures data from the ADC to the test data memory
void AD9467_Data_Interface::capture_test_data() {
	data_iface_core_->write(0x14, 1<<11);	
	wait_ms(1); 				// wait until test memory is filled
	data_iface_core_->write(0x14, 0<<11);	
}



// reads one sample of data from the test data memory
int AD9467_Data_Interface::read_test_data(int addr) {
	// set requested read address
	data_iface_core_->write(0x14, addr);

	// wait for read data ready (the read address equals the requested address)
	int addr_check;
	do {
		addr_check = data_iface_core_->read(0x18);
		addr_check = addr_check & 0x000007FF;
	}
	while (addr_check != addr);

	// read the data
	int read_data;
	read_data = data_iface_core_->read(0x18);
	read_data = read_data & 0x7FFF800;
	read_data = read_data>>11;

	return read_data;
}


// inits the delay controller (necessary for the delay controller)
int AD9467_Data_Interface::init_delay_controller() {

	// set and clear reset
	data_iface_core_->write(0x0c, 1);	// set reset bit
	wait_ms(10);
	data_iface_core_->write(0x0c, 0);	// clear reset after 10 ms


	// wait for reset to complete (wait for delay_ctrl_rdy set)
	int delay_ctrl_rdy;
	int timeout_counter = 0;

	do {
		delay_ctrl_rdy = data_iface_core_->read(0x10);
		wait_ms(10);
		timeout_counter++;
	}
	while ( (delay_ctrl_rdy == 0) && (timeout_counter < 100) );

	if (timeout_counter >= 100) return -1; 		// catch timeout

	return 0;
}

void AD9467_Data_Interface::check_availability_on_fpga() {
	cerr << "checking availability of ADC Data Interface IP Core core (checks if FPGA design is loaded).. ";
	int trash = data_iface_core_ -> read(0x08); // dummy read just to test if IP core responds
	cerr << " ok" << endl;
}