/*
  -------------------------------------------------------
    AD9467_SPI.cpp
    SPI interface driver class for the spi_ad_0 core for the AD9467
    Stefan Scholl, DC9ST, TU Kaiserslautern, 2014
  -------------------------------------------------------
*/

#include <iostream>
#include "AD9467_SPI.h"

using namespace std;


// private functions
	
// sends a single write instruction to the ADC
void AD9467_SPI::spi_write(int address, int data) {
	int spi_instruction = ((0x001FFF & address) << 8) | (0x000000FF & data);  // compose instruction of address and data (r/w bit and 2 w-bits = 0)

	spi_core_->write(0, spi_instruction);	// write instruction to core
	spi_core_->write(8, 0x00000001);        // start spi transaction 

	// wait for transaction to complete
	int timeout_counter = 1000000;

	while ( ((spi_core_->read(0x8) == 1) || (spi_core_->read(0xc) == 1)) && (timeout_counter > 0) ) { 	// check start bit and busy bit from the IP core for completion
		timeout_counter--;
	};

	if (timeout_counter <= 0) cerr << "AD9467 SPI: write instruction (A="<< address << ", D= " << data << ": TIMEOUT!" << endl;
};


// sends a simple read instruction to the ADC
int AD9467_SPI::spi_read(int address) {

	int spi_instruction = (0x00800000) | ((0x001FFF & address) << 8);  // compose instruction of address and data (r/w bit = 1 and 2 w-bits = 0)

	spi_core_->write(0, spi_instruction);	// write read instruction to core
	spi_core_->write(8, 0x00000001);		// start spi instruction

	// wait for instruction to complete
	int timeout_counter = 1000000;

	while ( ((spi_core_->read(0x8) == 1) || (spi_core_->read(0xc) == 1)) && (timeout_counter > 0) ) {
		timeout_counter--;
	};

	if (timeout_counter <= 0) cerr << "AD9467 SPI: read instruction: TIMEOUT!" << endl;

	return spi_core_->read(0x4);  // read received byte
};



// constructor & deconstructor

// class constructor
AD9467_SPI::AD9467_SPI(int spi_base_address, int size_in_k) {
	spi_core_ = new IP_Driver(spi_base_address, size_in_k);
}

// class deconstructor	
AD9467_SPI::~AD9467_SPI() {
	delete spi_core_;
}



// public functions

// resets ADC chip (soft reset)
void AD9467_SPI::reset_adc() {
	spi_write(0, 0x3C);   // soft reset (non buffered)

	do { }
	while (spi_read(0) != 24);	// wait unitl reset completes
}


// reads the chip ID (0x50 for the AD9467)
int AD9467_SPI::get_chip_id() {
	return spi_read(1);  // read chip ID from address 1	
}


// reads the chip grade (0x20 for the AD9467-250)
int AD9467_SPI::get_chip_grade() {
	return spi_read(2);
}


// reads the current test mode
int AD9467_SPI::get_test_mode() {
	return spi_read(0x0d);
}


// brief sets the test mode
int AD9467_SPI::set_test_mode(int test_mode) {

	if ((test_mode > 8) || (test_mode < 0))	return -1;

	int reg_content = spi_read(0x0d);
	reg_content = (reg_content & 0x30) | (test_mode & 0x0F);

	spi_write(0x0d, reg_content);
	spi_write(0xff, 1);

	return 0;
}

// set the output data format
int AD9467_SPI::set_data_format(int format) {

	if ((format > 2) || (format < 0))	return -1;

	int reg_content = spi_read(0x14);
	cerr << "set new data format, old format: " << hex << reg_content;
	reg_content = (reg_content & 0x1c) | (format & 0x03);

	spi_write(0x14, reg_content);
	spi_write(0xff, 1);

	cerr << ", new data format: " << hex << spi_read(0x14) << dec<< endl;

	return 0;
}